# Apply a different optimizer iteratively
from typing import Dict, Any, Sequence, Union
import os
from time import time
import random

import numpy as np
import torch

from .base import BaseOptimizer
from ..models import Model
from .lm import LM
from .. import config

__all__ = [
    "Iter",
    # "Iter_LM"
]


class Iter(BaseOptimizer):
    """Optimizer wrapper that performs optimization iteratively.

    This optimizer applies a different optimizer to a group model iteratively.
    It can be used for complex fits or when the number of models to fit is too large to fit in memory.

    Args:
        model: An `AstroPhot_Model` object to perform optimization on.
        method: The optimizer class to apply at each iteration step.
        initial_state: Optional initial state for optimization, defaults to None.
        max_iter: Maximum number of iterations, defaults to 100.
        method_kwargs: Keyword arguments to pass to `method`.
        **kwargs: Additional keyword arguments.

    Attributes:
        ndf: Degrees of freedom of the data.
        method: The optimizer class to apply at each iteration step. Default: Levenberg-Marquardt
        method_kwargs: Keyword arguments to pass to `method`.
        iteration: The number of iterations performed.
        lambda_history: A list of the states at each iteration step.
        loss_history: A list of the losses at each iteration step
    """

    def __init__(
        self,
        model: Model,
        initial_state: np.ndarray = None,
        max_iter: int = 100,
        lm_kwargs: Dict[str, Any] = {},
        **kwargs: Dict[str, Any],
    ) -> None:
        super().__init__(model, initial_state, max_iter=max_iter, **kwargs)

        self.current_state = model.build_params_array()
        self.lm_kwargs = lm_kwargs
        if "relative_tolerance" not in lm_kwargs:
            # Lower tolerance since it's not worth fine tuning a model when its neighbors will be shifting soon anyway
            self.lm_kwargs["relative_tolerance"] = 1e-3
            self.lm_kwargs["max_iter"] = 15
        #          # pixels      # parameters
        self.ndf = self.model.target[self.model.window].flatten("data").size(0) - len(
            self.current_state
        )
        if self.model.target.has_mask:
            # subtract masked pixels from degrees of freedom
            self.ndf -= torch.sum(self.model.target[self.model.window].flatten("mask")).item()

    def sub_step(self, model: Model, update_uncertainty=False) -> None:
        """
        Perform optimization for a single model.

        Args:
            model: The model to perform optimization on.
        """
        self.Y -= model()
        initial_values = model.target.copy()
        model.target = model.target - self.Y
        res = LM(model, **self.lm_kwargs).fit(update_uncertainty=update_uncertainty)
        self.Y += model()
        if self.verbose > 1:
            config.logger.info(res.message)
        model.target = initial_values

    def step(self) -> None:
        """
        Perform a single iteration of optimization.
        """
        if self.verbose > 0:
            config.logger.info("--------iter-------")

        # Fit each model individually
        for model in self.model.models:
            if self.verbose > 0:
                config.logger.info(model.name)
            self.sub_step(model)
        # Update the current state
        self.current_state = self.model.build_params_array()

        # Update the loss value
        with torch.no_grad():
            if self.verbose > 0:
                config.logger.info("Update Chi^2 with new parameters")
            self.Y = self.model(params=self.current_state)
            D = self.model.target[self.model.window].flatten("data")
            V = (
                self.model.target[self.model.window].flatten("variance")
                if self.model.target.has_variance
                else 1.0
            )
            if self.model.target.has_mask:
                M = self.model.target[self.model.window].flatten("mask")
                loss = (
                    torch.sum((((D - self.Y.flatten("data")) ** 2) / V)[torch.logical_not(M)])
                    / self.ndf
                )
            else:
                loss = torch.sum(((D - self.Y.flatten("data")) ** 2 / V)) / self.ndf
        if self.verbose > 0:
            config.logger.info(f"Loss: {loss.item()}")
        self.lambda_history.append(np.copy((self.current_state).detach().cpu().numpy()))
        self.loss_history.append(loss.item())

        # Test for convergence
        if self.iteration >= 2 and (
            (-self.relative_tolerance * 1e-3)
            < ((self.loss_history[-2] - self.loss_history[-1]) / self.loss_history[-1])
            < (self.relative_tolerance / 10)
        ):
            self._count_finish += 1
        else:
            self._count_finish = 0

        self.iteration += 1

    def fit(self) -> BaseOptimizer:
        self.iteration = 0
        self.Y = self.model(params=self.current_state)
        start_fit = time()
        try:
            while True:
                self.step()
                if self.iteration > 2 and self._count_finish >= 2:
                    self.message = self.message + "success"
                    break
                elif self.iteration >= self.max_iter:
                    self.message = self.message + f"fail max iterations reached: {self.iteration}"
                    break

        except KeyboardInterrupt:
            self.message = self.message + "fail interrupted"

        self.model.fill_dynamic_values(
            torch.tensor(self.res(), dtype=config.DTYPE, device=config.DEVICE)
        )
        if self.verbose > 1:
            config.logger.info(
                f"Iter Fitting complete in {time() - start_fit} sec with message: {self.message}"
            )

        return self


# class IterParam(BaseOptimizer):
#     """Optimization wrapper that call LM optimizer on subsets of variables.

#     IterParam takes the full set of parameters for a model and breaks
#     them down into chunks as specified by the user. It then calls
#     Levenberg-Marquardt optimization on the subset of parameters, and
#     iterates through all subsets until every parameter has been
#     optimized. It cycles through these chunks until convergence. This
#     method is very powerful in situations where the full optimization
#     problem cannot fit in memory, or where the optimization problem is
#     too complex to tackle as a single large problem. In full LM
#     optimization a single problematic parameter can ripple into issues
#     with every other parameter, so breaking the problem down can
#     sometimes make an otherwise intractable problem easier. For small
#     problems with only a few models, it is likely better to optimize
#     the full problem with LM as, when it works, LM is faster than the
#     IterParam method.

#     Args:
#       chunks (Union[int, tuple]): Specify how to break down the model parameters. If an integer, at each iteration the algorithm will break the parameters into groups of that size. If a tuple, should be a tuple of tuples of strings which give an explicit pairing of parameters to optimize, note that it is allowed to have variable size chunks this way. Default: 50
#       method (str): How to iterate through the chunks. Should be one of: random, sequential. Default: random
#     """

#     def __init__(
#         self,
#         model: Model,
#         initial_state: Sequence = None,
#         chunks: Union[int, tuple] = 50,
#         max_iter: int = 100,
#         method: str = "random",
#         LM_kwargs: dict = {},
#         **kwargs: Dict[str, Any],
#     ) -> None:
#         super().__init__(model, initial_state, max_iter=max_iter, **kwargs)

#         self.chunks = chunks
#         self.method = method
#         self.LM_kwargs = LM_kwargs

#         #          # pixels      # parameters
#         self.ndf = self.model.target[self.model.window].flatten("data").numel() - len(
#             self.current_state
#         )
#         if self.model.target.has_mask:
#             # subtract masked pixels from degrees of freedom
#             self.ndf -= torch.sum(self.model.target[self.model.window].flatten("mask")).item()

#     def step(self):
#         # These store the chunking information depending on which chunk mode is selected
#         param_ids = list(self.model.parameters.vector_identities())
#         init_param_ids = list(self.model.parameters.vector_identities())
#         _chunk_index = 0
#         _chunk_choices = None
#         res = None

#         if self.verbose > 0:
#             config.logger.info("--------iter-------")

#         # Loop through all the chunks
#         while True:
#             chunk = torch.zeros(len(init_param_ids), dtype=torch.bool, device=config.DEVICE)
#             if isinstance(self.chunks, int):
#                 if len(param_ids) == 0:
#                     break
#                 if self.method == "random":
#                     # Draw a random chunk of ids
#                     for pid in random.sample(param_ids, min(len(param_ids), self.chunks)):
#                         chunk[init_param_ids.index(pid)] = True
#                 else:
#                     # Draw the next chunk of ids
#                     for pid in param_ids[: self.chunks]:
#                         chunk[init_param_ids.index(pid)] = True
#                 # Remove the selected ids from the list
#                 for p in np.array(init_param_ids)[chunk.detach().cpu().numpy()]:
#                     param_ids.pop(param_ids.index(p))
#             elif isinstance(self.chunks, (tuple, list)):
#                 if _chunk_choices is None:
#                     # Make a list of the chunks as given explicitly
#                     _chunk_choices = list(range(len(self.chunks)))
#                 if self.method == "random":
#                     if len(_chunk_choices) == 0:
#                         break
#                     # Select a random chunk from the given groups
#                     sub_index = random.choice(_chunk_choices)
#                     _chunk_choices.pop(_chunk_choices.index(sub_index))
#                     for pid in self.chunks[sub_index]:
#                         chunk[param_ids.index(pid)] = True
#                 else:
#                     if _chunk_index >= len(self.chunks):
#                         break
#                     # Select the next chunk in order
#                     for pid in self.chunks[_chunk_index]:
#                         chunk[param_ids.index(pid)] = True
#                     _chunk_index += 1
#             else:
#                 raise ValueError(
#                     "Unrecognized chunks value, should be one of int, tuple. not: {type(self.chunks)}"
#                 )
#             if self.verbose > 1:
#                 config.logger.info(str(chunk))
#             del res
#             with Param_Mask(self.model.parameters, chunk):
#                 res = LM(
#                     self.model,
#                     ndf=self.ndf,
#                     **self.LM_kwargs,
#                 ).fit()
#             if self.verbose > 0:
#                 config.logger.info(f"chunk loss: {res.res_loss()}")
#             if self.verbose > 1:
#                 config.logger.info(f"chunk message: {res.message}")

#         self.loss_history.append(res.res_loss())
#         self.lambda_history.append(
#             self.model.parameters.vector_representation().detach().cpu().numpy()
#         )
#         if self.verbose > 0:
#             config.logger.info(f"Loss: {self.loss_history[-1]}")

#         # test for convergence
#         if self.iteration >= 2 and (
#             (-self.relative_tolerance * 1e-3)
#             < ((self.loss_history[-2] - self.loss_history[-1]) / self.loss_history[-1])
#             < (self.relative_tolerance / 10)
#         ):
#             self._count_finish += 1
#         else:
#             self._count_finish = 0

#         self.iteration += 1

#     def fit(self):
#         self.iteration = 0

#         start_fit = time()
#         try:
#             while True:
#                 self.step()
#                 if self.save_steps is not None:
#                     self.model.save(
#                         os.path.join(
#                             self.save_steps,
#                             f"{self.model.name}_Iteration_{self.iteration:03d}.yaml",
#                         )
#                     )
#                 if self.iteration > 2 and self._count_finish >= 2:
#                     self.message = self.message + "success"
#                     break
#                 elif self.iteration >= self.max_iter:
#                     self.message = self.message + f"fail max iterations reached: {self.iteration}"
#                     break

#         except KeyboardInterrupt:
#             self.message = self.message + "fail interrupted"

#         self.model.parameters.vector_set_representation(self.res())
#         if self.verbose > 1:
#             config.logger.info(
#                 f"Iter Fitting complete in {time() - start_fit} sec with message: {self.message}"
#             )

#         return self



from .image import (
    pixel_center_meshgrid,
    pixel_corner_meshgrid,
    pixel_simpsons_meshgrid,
    pixel_quad_meshgrid,
    rotate,
)
from .wcs import (
    world_to_plane_gnomonic,
    plane_to_world_gnomonic,
    pixel_to_plane_linear,
    plane_to_pixel_linear,
    sip_delta,
)
from .window import window_or, window_and

__all__ = (
    "pixel_center_meshgrid",
    "pixel_corner_meshgrid",
    "pixel_simpsons_meshgrid",
    "pixel_quad_meshgrid",
    "rotate",
    "world_to_plane_gnomonic",
    "plane_to_world_gnomonic",
    "pixel_to_plane_linear",
    "plane_to_pixel_linear",
    "sip_delta",
    "window_or",
    "window_and",
)



from caskade import Param as CParam
import torch


class Param(CParam):
    """
    A class that extends the Caskade Param class to include additional functionality.
    This class is used to define parameters for models in the AstroPhot package.
    """

    def __init__(self, *args, uncertainty=None, prof=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.uncertainty = uncertainty
        self.saveattrs.add("uncertainty")
        self.prof = prof
        self.saveattrs.add("prof")

    @property
    def uncertainty(self):
        return self._uncertainty

    @uncertainty.setter
    def uncertainty(self, uncertainty):
        if uncertainty is None:
            self._uncertainty = None
        else:
            self._uncertainty = torch.as_tensor(uncertainty)

    @property
    def prof(self):
        return self._prof

    @prof.setter
    def prof(self, prof):
        if prof is None:
            self._prof = None
        else:
            self._prof = torch.as_tensor(prof)

    @property
    def initialized(self):
        """Check if the parameter is initialized."""
        if self.pointer:
            return True
        if self.value is not None:
            return True
        return False

    def is_valid(self, value):
        if self.valid[0] is not None and torch.any(value <= self.valid[0]):
            return False
        if self.valid[1] is not None and torch.any(value >= self.valid[1]):
            return False
        return True

    def soft_valid(self, value):
        if self.valid[0] is None and self.valid[1] is None:
            return value
        if self.valid[0] is not None and self.valid[1] is not None:
            vrange = 0.1 * (self.valid[1] - self.valid[0])
            smin = self.valid[0] + 0.1 * vrange
            smax = self.valid[1] - 0.1 * vrange
        elif self.valid[0] is not None:
            smin = self.valid[0] + 0.1
            smax = None
        elif self.valid[1] is not None:
            smin = None
            smax = self.valid[1] - 0.1
        return torch.clamp(value, min=smin, max=smax)



import torch
from .sersic import sersic_n_to_b

b = sersic_n_to_b(1.0)


def exponential(R, Re, Ie):
    """Exponential 1d profile function, specifically designed for pytorch
    operations.

    Parameters:
        R: Radii tensor at which to evaluate the sersic function
        Re: Effective radius in the same units as R
        Ie: Effective surface density
    """
    return Ie * torch.exp(-b * ((R / Re) - 1.0))



import torch


def king(R, Rc, Rt, alpha, I0):
    """
    Empirical King profile.

    Parameters
    ----------
    R : array_like
        The radial distance from the center.
    Rc : float
        The core radius of the profile.
    Rt : float
        The truncation radius of the profile.
    alpha : float
        The power-law index of the profile.
    I0 : float
        The central intensity of the profile.

    Returns
    -------
    array_like
        The intensity at each radial distance.
    """
    beta = 1 / (1 + (Rt / Rc) ** 2) ** (1 / alpha)
    gamma = 1 / (1 + (R / Rc) ** 2) ** (1 / alpha)
    return torch.where(
        R < Rt, I0 * ((torch.clamp(gamma, 0, 1) - beta) / (1 - beta)) ** alpha, torch.zeros_like(R)
    )



import astrophot as ap
import torch
import numpy as np

from utils import make_basic_sersic
import pytest

######################################################################
# Image Objects
######################################################################


@pytest.fixture()
def sip_target():
    arr = torch.zeros((10, 15))
    return ap.SIPTargetImage(
        data=arr,
        pixelscale=1.0,
        zeropoint=1.0,
        variance=torch.ones_like(arr),
        mask=torch.zeros_like(arr),
        sipA={(1, 0): 1e-4, (0, 1): 1e-4, (2, 3): -1e-5},
        sipB={(1, 0): -1e-4, (0, 1): 5e-5, (2, 3): 2e-6},
        sipAP={(1, 0): -1e-4, (0, 1): -1e-4, (2, 3): 1e-5},
        sipBP={(1, 0): 1e-4, (0, 1): -5e-5, (2, 3): -2e-6},
    )


def test_sip_image_creation(sip_target):
    assert sip_target.pixelscale == 1.0, "image should track pixelscale"
    assert sip_target.zeropoint == 1.0, "image should track zeropoint"
    assert sip_target.crpix[0] == 0, "image should track crpix"
    assert sip_target.crpix[1] == 0, "image should track crpix"

    slicer = ap.Window((7, 13, 4, 7), sip_target)
    sliced_image = sip_target[slicer]
    assert sliced_image.crpix[0] == -7, "crpix of subimage should give relative position"
    assert sliced_image.crpix[1] == -4, "crpix of subimage should give relative position"
    assert sliced_image.shape == (6, 3), "sliced image should have correct shape"
    assert sliced_image.pixel_area_map.shape == (
        6,
        3,
    ), "sliced image should have correct pixel area map shape"
    assert sliced_image.distortion_ij.shape == (
        2,
        6,
        3,
    ), "sliced image should have correct distortion shape"
    assert sliced_image.distortion_IJ.shape == (
        2,
        6,
        3,
    ), "sliced image should have correct distortion shape"

    sip_model_image = sip_target.model_image(upsample=2, pad=1)
    assert sip_model_image.shape == (32, 22), "model image should have correct shape"
    assert sip_model_image.pixel_area_map.shape == (
        32,
        22,
    ), "model image pixel area map should have correct shape"
    assert sip_model_image.distortion_ij.shape == (
        2,
        32,
        22,
    ), "model image distortion model should have correct shape"
    assert sip_model_image.distortion_IJ.shape == (
        2,
        32,
        22,
    ), "model image distortion model should have correct shape"

    # reduce
    sip_model_reduce = sip_model_image.reduce(scale=1)
    assert sip_model_reduce is sip_model_image, "reduce should return the same image if scale is 1"
    sip_model_reduce = sip_model_image.reduce(scale=2)
    assert sip_model_reduce.shape == (16, 11), "reduced model image should have correct shape"

    # crop
    sip_model_crop = sip_model_image.crop(1)
    assert sip_model_crop.shape == (30, 20), "cropped model image should have correct shape"
    sip_model_crop = sip_model_image.crop([1])
    assert sip_model_crop.shape == (30, 20), "cropped model image should have correct shape"
    sip_model_crop = sip_model_image.crop([1, 2])
    assert sip_model_crop.shape == (30, 18), "cropped model image should have correct shape"
    sip_model_crop = sip_model_image.crop([1, 2, 3, 4])
    assert sip_model_crop.shape == (29, 15), "cropped model image should have correct shape"

    sip_model_crop.fluxdensity_to_flux()
    assert torch.all(
        sip_model_crop.data >= 0
    ), "cropped model image data should be non-negative after flux density to flux conversion"


def test_sip_image_wcs_roundtrip(sip_target):
    """
    Test that the WCS roundtrip works correctly for SIP images.
    """
    i, j = sip_target.pixel_center_meshgrid()
    x, y = sip_target.pixel_to_plane(i, j)
    i2, j2 = sip_target.plane_to_pixel(x, y)

    assert torch.allclose(i, i2, atol=0.5), "i coordinates should match after WCS roundtrip"
    assert torch.allclose(j, j2, atol=0.5), "j coordinates should match after WCS roundtrip"


def test_sip_image_save_load(sip_target):
    """
    Test that SIP images can be saved and loaded correctly.
    """
    # Save the SIP image to a file
    sip_target.save("test_sip_image.fits")

    # Load the SIP image from the file
    loaded_image = ap.SIPTargetImage(filename="test_sip_image.fits")

    # Check that the loaded image matches the original
    assert torch.allclose(
        sip_target.data, loaded_image.data
    ), "Loaded image data should match original"
    assert torch.allclose(
        sip_target.pixelscale, loaded_image.pixelscale
    ), "Loaded image pixelscale should match original"
    assert torch.allclose(
        sip_target.zeropoint, loaded_image.zeropoint
    ), "Loaded image zeropoint should match original"
    print(loaded_image.sipA)
    assert all(
        np.allclose(sip_target.sipA[key], loaded_image.sipA[key]) for key in sip_target.sipA
    ), "Loaded image sipA should match original"
    assert all(
        np.allclose(sip_target.sipB[key], loaded_image.sipB[key]) for key in sip_target.sipB
    ), "Loaded image sipB should match original"
    assert all(
        np.allclose(sip_target.sipAP[key], loaded_image.sipAP[key]) for key in sip_target.sipAP
    ), "Loaded image sipAP should match original"
    assert all(
        np.allclose(sip_target.sipBP[key], loaded_image.sipBP[key]) for key in sip_target.sipBP
    ), "Loaded image sipBP should match original"



from typing import Sequence, Optional

import numpy as np
import torch
from scipy.optimize import minimize
from scipy.special import gammainc

from .. import config
from ..models import Model
from ..image import Window
from ..param import ValidContext


__all__ = ["BaseOptimizer"]


class BaseOptimizer(object):
    """
    Base optimizer object that other optimizers inherit from. Ensures consistent signature for the classes.

    Parameters:
        model: an AstroPhot_Model object that will have its (unlocked) parameters optimized [AstroPhot_Model]
        initial_state: optional initialization for the parameters as a 1D tensor [tensor]
        max_iter: maximum allowed number of iterations [int]
        relative_tolerance: tolerance for counting success steps as: 0 < (Chi2^2 - Chi1^2)/Chi1^2 < tol [float]

    """

    def __init__(
        self,
        model: Model,
        initial_state: Sequence = None,
        relative_tolerance: float = 1e-3,
        fit_window: Optional[Window] = None,
        verbose: int = 0,
        max_iter: int = None,
        save_steps: Optional[str] = None,
        fit_valid: bool = True,
    ) -> None:
        """
        Initializes a new instance of the class.

        Args:
            model (object): An object representing the model.
            initial_state (Optional[Sequence]): The initial state of the model could be any tensor.
                           If `None`, the model's default initial state will be used.
            relative_tolerance (float): The relative tolerance for the optimization.
            fit_parameters_identity (Optional[tuple]): a tuple of parameter identity strings which tell the LM optimizer which parameters of the model to fit.
            **kwargs (dict): Additional keyword arguments.

        Attributes:
            model (object): An object representing the model.
            verbose (int): The verbosity level.
            current_state (Tensor): The current state of the model.
            max_iter (int): The maximum number of iterations.
            iteration (int): The current iteration number.
            save_steps (Optional[str]): Save intermediate results to this path.
            relative_tolerance (float): The relative tolerance for the optimization.
            lambda_history (List[ndarray]): A list of the optimization steps.
            loss_history (List[float]): A list of the optimization losses.
            message (str): An informational message.
        """

        self.model = model
        self.verbose = verbose

        if initial_state is None:
            self.current_state = model.build_params_array()
        else:
            self.current_state = torch.as_tensor(
                initial_state, dtype=model.dtype, device=model.device
            )

        if fit_window is None:
            self.fit_window = self.model.window
        else:
            self.fit_window = fit_window & self.model.window

        self.max_iter = max_iter if max_iter is not None else 100 * len(self.current_state)
        self.iteration = 0
        self.save_steps = save_steps
        self.fit_valid = fit_valid

        self.relative_tolerance = relative_tolerance
        self.lambda_history = []
        self.loss_history = []
        self.message = ""

    def fit(self) -> "BaseOptimizer":
        """
        Raises:
            NotImplementedError: Error is raised if this method is not implemented in a subclass of BaseOptimizer.
        """
        raise NotImplementedError("Please use a subclass of BaseOptimizer for optimization")

    def step(self, current_state: torch.Tensor = None) -> None:
        """Args:
            current_state (torch.Tensor, optional): Current state of the model parameters. Defaults to None.

        Raises:
            NotImplementedError: Error is raised if this method is not implemented in a subclass of BaseOptimizer.
        """
        raise NotImplementedError("Please use a subclass of BaseOptimizer for optimization")

    def chi2min(self) -> float:
        """
        Returns the minimum value of chi^2 loss in the loss history.

        Returns:
                float: Minimum value of chi^2 loss.
        """
        return np.nanmin(self.loss_history)

    def res(self) -> np.ndarray:
        """Returns the value of lambda (regularization strength) at which minimum chi^2 loss was achieved.

        Returns: ndarray which is the Value of lambda at which minimum chi^2 loss was achieved.
        """
        N = np.isfinite(self.loss_history)
        if np.sum(N) == 0:
            config.logger.warning(
                "Getting optimizer res with no real loss history, using current state"
            )
            return self.current_state.detach().cpu().numpy()
        return np.array(self.lambda_history)[N][np.argmin(np.array(self.loss_history)[N])]

    def res_loss(self):
        N = np.isfinite(self.loss_history)
        return np.min(np.array(self.loss_history)[N])

    @staticmethod
    def chi2contour(n_params: int, confidence: float = 0.682689492137) -> float:
        """
        Calculates the chi^2 contour for the given number of parameters.

        Args:
            n_params (int): The number of parameters.
            confidence (float, optional): The confidence interval (default is 0.682689492137).

        Returns:
            float: The calculated chi^2 contour value.

        Raises:
            RuntimeError: If unable to compute the Chi^2 contour for the given number of parameters.

        """

        def _f(x: float, nu: int) -> float:
            """Helper function for calculating chi^2 contour."""
            return (gammainc(nu / 2, x / 2) - confidence) ** 2

        for method in ["L-BFGS-B", "Powell", "Nelder-Mead"]:
            res = minimize(_f, x0=n_params, args=(n_params,), method=method, tol=1e-8)

            if res.success:
                return res.x[0]
        raise RuntimeError(f"Unable to compute Chi^2 contour for n params: {n_params}")



from typing import Union, Tuple

import numpy as np

from ..errors import InvalidWindow

__all__ = ("Window",)


class Window:
    def __init__(
        self,
        window: Union[Tuple[int, int, int, int], Tuple[Tuple[int, int], Tuple[int, int]]],
        image: "Image",
    ):
        self.extent = window
        self.image = image

    @property
    def identity(self):
        return self.image.identity

    @property
    def crpix(self):
        return self.image.crpix

    @property
    def shape(self):
        return (self.i_high - self.i_low, self.j_high - self.j_low)

    @property
    def extent(self):
        return (self.i_low, self.i_high, self.j_low, self.j_high)

    @extent.setter
    def extent(
        self, value: Union[Tuple[int, int, int, int], Tuple[Tuple[int, int], Tuple[int, int]]]
    ):
        if len(value) == 4:
            self.i_low, self.i_high, self.j_low, self.j_high = value
        elif len(value) == 2:
            self.i_low, self.j_low = value[0]
            self.i_high, self.j_high = value[1]
        else:
            raise ValueError(
                "Extent must be formatted as (i_low, i_high, j_low, j_high) or ((i_low, j_low), (i_high, j_high))"
            )

    def chunk(self, chunk_size: int):
        # number of pixels on each axis
        px = self.i_high - self.i_low
        py = self.j_high - self.j_low
        # total number of chunks desired
        chunk_tot = int(np.ceil((px * py) / chunk_size))
        # number of chunks on each axis
        cx = int(np.ceil(np.sqrt(chunk_tot * px / py)))
        cy = int(np.ceil(chunk_tot / cx))
        # number of pixels on each axis per chunk
        stepx = int(np.ceil(px / cx))
        stepy = int(np.ceil(py / cy))
        # create the windows
        windows = []
        for i in range(self.i_low, self.i_high, stepx):
            for j in range(self.j_low, self.j_high, stepy):
                i_high = min(i + stepx, self.i_high)
                j_high = min(j + stepy, self.j_high)
                windows.append(Window((i, i_high, j, j_high), self.image))
        return windows

    def pad(self, pad: int):
        self.i_low -= pad
        self.i_high += pad
        self.j_low -= pad
        self.j_high += pad

    def copy(self):
        return Window((self.i_low, self.i_high, self.j_low, self.j_high), self.image)

    def __or__(self, other: "Window"):
        if not isinstance(other, Window):
            raise TypeError(f"Cannot combine Window with {type(other)}")
        if self.image != other.image:
            raise InvalidWindow(
                f"Cannot combine Windows from different images: {self.image.identity} and {other.image.identity}"
            )
        new_i_low = min(self.i_low, other.i_low)
        new_i_high = max(self.i_high, other.i_high)
        new_j_low = min(self.j_low, other.j_low)
        new_j_high = max(self.j_high, other.j_high)
        return Window((new_i_low, new_i_high, new_j_low, new_j_high), self.image)

    def __ior__(self, other: "Window"):
        if not isinstance(other, Window):
            raise TypeError(f"Cannot combine Window with {type(other)}")
        if self.image != other.image:
            raise InvalidWindow(
                f"Cannot combine Windows from different images: {self.image.identity} and {other.image.identity}"
            )
        self.i_low = min(self.i_low, other.i_low)
        self.i_high = max(self.i_high, other.i_high)
        self.j_low = min(self.j_low, other.j_low)
        self.j_high = max(self.j_high, other.j_high)
        return self

    def __and__(self, other: "Window"):
        if not isinstance(other, Window):
            raise TypeError(f"Cannot intersect Window with {type(other)}")
        if self.image.identity != other.image.identity:
            raise InvalidWindow(
                f"Cannot combine Windows from different images: {self.image.identity} and {other.image.identity}"
            )
        if (
            self.i_high <= other.i_low
            or self.i_low >= other.i_high
            or self.j_high <= other.j_low
            or self.j_low >= other.j_high
        ):
            return Window((0, 0, 0, 0), self.image)
        # fixme handle crpix
        new_i_low = max(self.i_low, other.i_low)
        new_i_high = min(self.i_high, other.i_high)
        new_j_low = max(self.j_low, other.j_low)
        new_j_high = min(self.j_high, other.j_high)
        return Window((new_i_low, new_i_high, new_j_low, new_j_high), self.image)

    def __str__(self):
        return f"Window({self.i_low}, {self.i_high}, {self.j_low}, {self.j_high})"


class WindowList:
    def __init__(self, windows: list[Window]):
        if not all(isinstance(window, Window) for window in windows):
            raise InvalidWindow(
                f"Window_List can only hold Window objects, not {tuple(type(window) for window in windows)}"
            )
        self.windows = windows

    def index(self, other: Window):
        for i, window in enumerate(self.windows):
            if other.identity == window.identity:
                return i
        else:
            raise IndexError("Could not find identity match between window list and input window")

    def __and__(self, other: "WindowList"):
        if not isinstance(other, WindowList):
            raise TypeError(f"Cannot intersect WindowList with {type(other)}")
        if len(self.windows) == 0 or len(other.windows) == 0:
            return WindowList([])
        new_windows = []
        for other_window in other.windows:
            try:
                i = self.index(other_window)
            except IndexError:
                continue  # skip if the window is not in self.windows
            new_windows.append(self.windows[i] & other_window)
        return WindowList(new_windows)

    def __getitem__(self, index):
        return self.windows[index]

    def __len__(self):
        return len(self.windows)

    def __iter__(self):
        return iter(self.windows)



import numpy as np

deg_to_arcsec = 3600.0
arcsec_to_deg = 1.0 / deg_to_arcsec


def flux_to_sb(flux, pixel_area, zeropoint):
    """Conversion from flux units to logarithmic surface brightness
    units.

    .. math::

      \\mu = -2.5\\log_{10}(flux) + z.p. + 2.5\\log_{10}(A)

    where :math:`z.p.` is the zeropoint and :math:`A` is the area of a pixel.

    """
    return -2.5 * np.log10(flux) + zeropoint + 2.5 * np.log10(pixel_area)


def flux_to_mag(flux, zeropoint, fluxe=None):
    """Converts a flux total into logarithmic magnitude units.

    .. math::

      m = -2.5\\log_{10}(flux) + z.p.

    where :math:`z.p.` is the zeropoint.

    """
    if fluxe is None:
        return -2.5 * np.log10(flux) + zeropoint
    else:
        return -2.5 * np.log10(flux) + zeropoint, 2.5 * fluxe / (np.log(10) * flux)


def sb_to_flux(sb, pixel_area, zeropoint):
    """Converts logarithmic surface brightness units into flux units.

    .. math::

      flux = A 10^{-(\\mu - z.p.)/2.5}

    where :math:`z.p.` is the zeropoint and :math:`A` is the area of a pixel.

    """
    return pixel_area * 10 ** (-(sb - zeropoint) / 2.5)


def mag_to_flux(mag, zeropoint, mage=None):
    """converts logarithmic magnitude units into a flux total.

    .. math::

      flux = 10^{-(m - z.p.)/2.5}

    where :math:`z.p.` is the zeropoint.

    """
    if mage is None:
        return 10 ** (-(mag - zeropoint) / 2.5)
    else:
        I = 10 ** (-(mag - zeropoint) / 2.5)
        return I, np.log(10) * I * mage / 2.5


def magperarcsec2_to_mag(mu, a=None, b=None, A=None):
    """
    Converts mag/arcsec^2 to mag
    mu: mag/arcsec^2
    a: semi major axis radius (arcsec)
    b: semi minor axis radius (arcsec)
    A: pre-calculated area (arcsec^2)
    returns: mag


    .. math::

      m = \\mu -2.5\\log_{10}(A)

    where :math:`A` is an area in arcsec^2.

    """
    assert (A is not None) or (a is not None and b is not None)
    if A is None:
        A = np.pi * a * b
    return mu - 2.5 * np.log10(
        A
    )  # https://en.wikipedia.org/wiki/Surface_brightness#Calculating_surface_brightness


def mag_to_magperarcsec2(m, a=None, b=None, R=None, A=None):
    """
    Converts mag to mag/arcsec^2
    m: mag
    a: semi major axis radius (arcsec)
    b: semi minor axis radius (arcsec)
    A: pre-calculated area (arcsec^2)
    returns: mag/arcsec^2

    .. math::

      \\mu = m + 2.5\\log_{10}(A)

    where :math:`A` is an area in arcsec^2.
    """
    assert (A is not None) or (a is not None and b is not None) or (R is not None)
    if R is not None:
        A = np.pi * (R**2)
    elif A is None:
        A = np.pi * a * b
    return m + 2.5 * np.log10(
        A
    )  # https://en.wikipedia.org/wiki/Surface_brightness#Calculating_surface_brightness


def PA_shift_convention(pa, unit="rad"):
    """
    Alternates between standard mathematical convention for angles, and astronomical position angle convention.
    The standard convention is to measure angles counter-clockwise relative to the positive x-axis
    The astronomical convention is to measure angles counter-clockwise relative to the positive y-axis
    """

    if unit == "rad":
        shift = np.pi
    elif unit == "deg":
        shift = 180.0

    return (pa - (shift / 2)) % shift



import numpy as np
import torch

deg_to_rad = np.pi / 180
rad_to_deg = 180 / np.pi
rad_to_arcsec = rad_to_deg * 3600
arcsec_to_rad = deg_to_rad / 3600


def world_to_plane_gnomonic(ra, dec, ra0, dec0, x0=0.0, y0=0.0):
    """
    Convert world coordinates (RA, Dec) to plane coordinates (x, y) using the gnomonic projection.

    Parameters
    ----------
    ra : torch.Tensor
        Right Ascension in degrees.
    dec : torch.Tensor
        Declination in degrees.
    ra0 : torch.Tensor
        Reference Right Ascension in degrees.
    dec0 : torch.Tensor
        Reference Declination in degrees.

    Returns
    -------
    x : torch.Tensor
        x coordinate in arcseconds.
    y : torch.Tensor
        y coordinate in arcseconds.
    """
    ra = ra * deg_to_rad
    dec = dec * deg_to_rad
    ra0 = ra0 * deg_to_rad
    dec0 = dec0 * deg_to_rad

    cosc = torch.sin(dec0) * torch.sin(dec) + torch.cos(dec0) * torch.cos(dec) * torch.cos(ra - ra0)

    x = torch.cos(dec) * torch.sin(ra - ra0)

    y = torch.cos(dec0) * torch.sin(dec) - torch.sin(dec0) * torch.cos(dec) * torch.cos(ra - ra0)

    return x * rad_to_arcsec / cosc + x0, y * rad_to_arcsec / cosc + y0


def plane_to_world_gnomonic(x, y, ra0, dec0, x0=0.0, y0=0.0, s=1e-10):
    """
    Convert plane coordinates (x, y) to world coordinates (RA, Dec) using the gnomonic projection.
    Parameters
    ----------
    x : torch.Tensor
        x coordinate in arcseconds.
    y : torch.Tensor
        y coordinate in arcseconds.
    ra0 : torch.Tensor
        Reference Right Ascension in degrees.
    dec0 : torch.Tensor
        Reference Declination in degrees.
    s : float
        Small constant to avoid division by zero.

    Returns
    -------
    ra : torch.Tensor
        Right Ascension in degrees.
    dec : torch.Tensor
        Declination in degrees.
    """
    x = (x - x0) * arcsec_to_rad
    y = (y - y0) * arcsec_to_rad
    ra0 = ra0 * deg_to_rad
    dec0 = dec0 * deg_to_rad

    rho = torch.sqrt(x**2 + y**2) + s
    c = torch.arctan(rho)

    ra = ra0 + torch.arctan2(
        x * torch.sin(c),
        rho * torch.cos(dec0) * torch.cos(c) - y * torch.sin(dec0) * torch.sin(c),
    )

    dec = torch.arcsin(torch.cos(c) * torch.sin(dec0) + y * torch.sin(c) * torch.cos(dec0) / rho)

    return ra * rad_to_deg, dec * rad_to_deg


def pixel_to_plane_linear(i, j, i0, j0, CD, x0=0.0, y0=0.0):
    """
    Convert pixel coordinates to a tangent plane using the WCS information. This
    matches the FITS convention for linear transformations.

    Parameters
    ----------
    i: Tensor
        The first coordinate of the pixel in pixel units.
    j: Tensor
        The second coordinate of the pixel in pixel units.
    i0: Tensor
        The i reference pixel coordinate in pixel units.
    j0: Tensor
        The j reference pixel coordinate in pixel units.
    CD: Tensor
        The CD matrix in arcsec per pixel. This 2x2 matrix is used to convert
        from pixel to arcsec units and also handles rotation/skew.
    x0: float
        The x reference coordinate in arcsec.
    y0: float
        The y reference coordinate in arcsec.

    Returns
    -------
    Tuple: [Tensor, Tensor]
        Tuple containing the x and y tangent plane coordinates in arcsec.
    """
    uv = torch.stack((i.flatten() - i0, j.flatten() - j0), dim=0)
    xy = CD @ uv

    return xy[0].reshape(i.shape) + x0, xy[1].reshape(i.shape) + y0


def sip_delta(u, v, sipA=(), sipB=()):
    """
    u = j - j0
    v = i - i0
    sipA = dict(tuple(int,int), float)
        The SIP coefficients, where the keys are tuples of powers (i, j) and the values are the coefficients.
        For example, {(1, 2): 0.1} means delta_u = 0.1 * (u * v^2).
    """
    delta_u = torch.zeros_like(u)
    delta_v = torch.zeros_like(v)
    # Get all used coefficient powers
    all_a = set(s[0] for s in sipA) | set(s[0] for s in sipB)
    all_b = set(s[1] for s in sipA) | set(s[1] for s in sipB)
    # Pre-compute all powers of u and v
    u_a = dict((a, u**a) for a in all_a)
    v_b = dict((b, v**b) for b in all_b)
    for a, b in sipA:
        delta_u = delta_u + sipA[(a, b)] * (u_a[a] * v_b[b])
    for a, b in sipB:
        delta_v = delta_v + sipB[(a, b)] * (u_a[a] * v_b[b])
    return delta_u, delta_v


def pixel_to_plane_sip(i, j, i0, j0, CD, sip_powers=[], sip_coefs=[], x0=0.0, y0=0.0):
    """
    Convert pixel coordinates to a tangent plane using the WCS information. This
    matches the FITS convention for SIP transformations.

    For more information see:

    * FITS World Coordinate System (WCS):
      https://fits.gsfc.nasa.gov/fits_wcs.html
    * Representations of world coordinates in FITS, 2002, by Geisen and
      Calabretta
    * The SIP Convention for Representing Distortion in FITS Image Headers,
      2008, by Shupe and Hook

    Parameters
    ----------
    i: Tensor
        The first coordinate of the pixel in pixel units.
    j: Tensor
        The second coordinate of the pixel in pixel units.
    i0: Tensor
        The i reference pixel coordinate in pixel units.
    j0: Tensor
        The j reference pixel coordinate in pixel units.
    CD: Tensor
        The CD matrix in degrees per pixel. This 2x2 matrix is used to convert
        from pixel to degree units and also handles rotation/skew.
    sip_powers: Tensor
        The powers of the pixel coordinates for the SIP distortion, should be a
        shape (N orders, 2) tensor. ``N orders`` is the number of non-zero
        polynomial coefficients. The second axis has the powers in order ``i,
        j``.
    sip_coefs: Tensor
        The coefficients of the pixel coordinates for the SIP distortion, should
        be a shape (N orders, 2) tensor. ``N orders`` is the number of non-zero
        polynomial coefficients. The second axis has the coefficients in order
        ``delta_x, delta_y``.
    x0: float
        The x reference coordinate in arcsec.
    y0: float
        The y reference coordinate in arcsec.

    Note
    ----
    The representation of the SIP powers and coefficients assumes that the SIP
    polynomial will use the same orders for both the x and y coordinates. If
    this is not the case you may use zeros for the coefficients to ensure all
    polynomial combinations are evaluated. However, it is very common to have
    the same orders for both.

    Returns
    -------
    Tuple: [Tensor, Tensor]
        Tuple containing the x and y tangent plane coordinates in arcsec.
    """
    uv = torch.stack((j.reshape(-1) - j0, i.reshape(-1) - i0), dim=1)
    delta_p = torch.zeros_like(uv)
    for p in range(len(sip_powers)):
        delta_p += sip_coefs[p] * torch.prod(uv ** sip_powers[p], dim=-1).unsqueeze(-1)
    plane = torch.einsum("ij,...j->...i", CD, uv + delta_p)
    return plane[..., 0] + x0, plane[..., 1] + y0


def plane_to_pixel_linear(x, y, i0, j0, CD, x0=0.0, y0=0.0):
    """
    Convert tangent plane coordinates to pixel coordinates using the WCS
    information. This matches the FITS convention for linear transformations.

    Parameters
    ----------
    x: Tensor
        The first coordinate of the pixel in arcsec.
    y: Tensor
        The second coordinate of the pixel in arcsec.
    i0: Tensor
        The i reference pixel coordinate in pixel units.
    j0: Tensor
        The j reference pixel coordinate in pixel units.
    iCD: Tensor
        The inverse CD matrix in arcsec per pixel. This 2x2 matrix is used to convert
        from pixel to arcsec units and also handles rotation/skew.
    x0: float
        The x reference coordinate in arcsec.
    y0: float
        The y reference coordinate in arcsec.

    Returns
    -------
    Tuple: [Tensor, Tensor]
        Tuple containing the i and j pixel coordinates in pixel units.
    """
    xy = torch.stack((x.flatten() - x0, y.flatten() - y0), dim=0)
    uv = torch.linalg.inv(CD) @ xy

    return uv[0].reshape(x.shape) + i0, uv[1].reshape(y.shape) + j0



from functools import wraps
import warnings

import numpy as np


class classproperty:
    def __init__(self, fget):
        self.fget = fget

    def __get__(self, instance, owner):
        return self.fget(owner)


def ignore_numpy_warnings(func):
    """This decorator is used to turn off numpy warnings. This should
    only be used in initialize scripts which often run heuristic code
    to determine initial parameter values. These heuristics may
    encounter log(0) or sqrt(-1) or other numerical artifacts and
    should handle them before returning. This decorator simply cleans
    up that processes to minimize clutter in the output.

    """

    @wraps(func)
    def wrapped(*args, **kwargs):
        old_settings = np.seterr(all="ignore")
        warnings.filterwarnings("ignore", category=np.exceptions.VisibleDeprecationWarning)
        result = func(*args, **kwargs)
        np.seterr(**old_settings)
        warnings.filterwarnings("default", category=np.exceptions.VisibleDeprecationWarning)
        return result

    return wrapped


def combine_docstrings(cls):
    combined_docs = [cls.__doc__ or ""]
    for base in cls.__bases__:
        if base.__doc__:
            combined_docs.append(f"\n[UNIT {base.__name__}]\n{base.__doc__}")
    cls.__doc__ = "\n".join(combined_docs).strip()
    return cls



# Hamiltonian Monte-Carlo
from typing import Optional, Sequence

import torch

try:
    import pyro
    import pyro.distributions as dist
    from pyro.infer import MCMC as pyro_MCMC
    from pyro.infer import HMC as pyro_HMC
    from pyro.infer.mcmc.adaptation import BlockMassMatrix
    from pyro.ops.welford import WelfordCovariance
except ImportError:
    pyro = None

from .base import BaseOptimizer
from ..models import Model

__all__ = ["HMC"]


###########################################
# !Overwrite pyro configuration behavior!
# currently this is the only way to provide
# mass matrix manually
###########################################
def new_configure(self, mass_matrix_shape, adapt_mass_matrix=True, options={}):
    """
    Sets up an initial mass matrix.

    :param dict mass_matrix_shape: a dict that maps tuples of site names to the shape of
        the corresponding mass matrix. Each tuple of site names corresponds to a block.
    :param bool adapt_mass_matrix: a flag to decide whether an adaptation scheme will be used.
    :param dict options: tensor options to construct the initial mass matrix.
    """
    inverse_mass_matrix = {}
    for site_names, shape in mass_matrix_shape.items():
        self._mass_matrix_size[site_names] = shape[0]
        diagonal = len(shape) == 1
        inverse_mass_matrix[site_names] = (
            torch.full(shape, self._init_scale, **options)
            if diagonal
            else torch.eye(*shape, **options) * self._init_scale
        )
        if adapt_mass_matrix:
            adapt_scheme = WelfordCovariance(diagonal=diagonal)
            self._adapt_scheme[site_names] = adapt_scheme

    if len(self.inverse_mass_matrix.keys()) == 0:
        self.inverse_mass_matrix = inverse_mass_matrix


BlockMassMatrix.configure = new_configure
############################################


class HMC(BaseOptimizer):
    """Hamiltonian Monte-Carlo sampler wrapper for the Pyro package.

    This MCMC algorithm uses gradients of the Chi^2 to more
    efficiently explore the probability distribution. Consider using
    the NUTS sampler instead of HMC, as it is generally better in most
    aspects.

    More information on HMC can be found at:
    https://en.wikipedia.org/wiki/Hamiltonian_Monte_Carlo,
    https://arxiv.org/abs/1701.02434, and
    http://www.mcmchandbook.net/HandbookChapter5.pdf

    Args:
        model (AstroPhot_Model): The model which will be sampled.
        initial_state (Optional[Sequence], optional): A 1D array with the values for each parameter in the model. These values should be in the form of "as_representation" in the model. Defaults to None.
        max_iter (int, optional): The number of sampling steps to perform. Defaults to 1000.
        epsilon (float, optional): The length of the integration step to perform for each leapfrog iteration. The momentum update will be of order epsilon * score. Defaults to 1e-5.
        leapfrog_steps (int, optional): Number of steps to perform with leapfrog integrator per sample of the HMC. Defaults to 20.
        inv_mass (float or array, optional): Inverse Mass matrix (covariance matrix) which can tune the behavior in each dimension to ensure better mixing when sampling. Defaults to the identity.
        progress_bar (bool, optional): Whether to display a progress bar during sampling. Defaults to True.
        prior (distribution, optional): Prior distribution for the parameters. Defaults to None.
        warmup (int, optional): Number of warmup steps before actual sampling begins. Defaults to 100.
        hmc_kwargs (dict, optional): Additional keyword arguments for the HMC sampler. Defaults to {}.
        mcmc_kwargs (dict, optional): Additional keyword arguments for the MCMC process. Defaults to {}.

    """

    def __init__(
        self,
        model: Model,
        initial_state: Optional[Sequence] = None,
        max_iter: int = 1000,
        inv_mass: Optional[torch.Tensor] = None,
        epsilon: float = 1e-5,
        leapfrog_steps: int = 20,
        progress_bar: bool = True,
        prior: Optional[dist.Distribution] = None,
        warmup: int = 100,
        hmc_kwargs: dict = {},
        mcmc_kwargs: dict = {},
        likelihood: str = "gaussian",
        **kwargs,
    ):
        if pyro is None:
            raise ImportError("Pyro must be installed to use HMC.")
        super().__init__(model, initial_state, max_iter=max_iter, **kwargs)

        self.inv_mass = inv_mass
        self.epsilon = epsilon
        self.leapfrog_steps = leapfrog_steps
        self.progress_bar = progress_bar
        self.prior = prior
        self.warmup = warmup
        self.hmc_kwargs = hmc_kwargs
        self.mcmc_kwargs = mcmc_kwargs
        self.likelihood = likelihood
        self.acceptance = None

    def fit(
        self,
        state: Optional[torch.Tensor] = None,
    ):
        """Performs MCMC sampling using Hamiltonian Monte-Carlo step.

        Records the chain for later examination.

        Args:
            state (torch.Tensor, optional): Model parameters as a 1D tensor.

        Returns:
            HMC: An instance of the HMC class with updated chain.

        """

        def step(model, prior):
            x = pyro.sample("x", prior)
            # Log-likelihood function
            if self.likelihood == "gaussian":
                log_likelihood_value = model.gaussian_log_likelihood(params=x)
            elif self.likelihood == "poisson":
                log_likelihood_value = model.poisson_log_likelihood(params=x)
            else:
                raise ValueError(f"Unsupported likelihood type: {self.likelihood}")
            # Observe the log-likelihood
            pyro.factor("obs", log_likelihood_value)

        if self.prior is None:
            self.prior = dist.Normal(
                self.current_state,
                torch.ones_like(self.current_state) * 1e2 + torch.abs(self.current_state) * 1e2,
            )

        # Set up the HMC sampler
        hmc_kwargs = {
            "jit_compile": False,
            "ignore_jit_warnings": True,
            "full_mass": True,
            "step_size": self.epsilon,
            "num_steps": self.leapfrog_steps,
            "adapt_step_size": False,
            "adapt_mass_matrix": self.inv_mass is None,
        }
        hmc_kwargs.update(self.hmc_kwargs)
        hmc_kernel = pyro_HMC(step, **hmc_kwargs)
        if self.inv_mass is not None:
            hmc_kernel.mass_matrix_adapter.inverse_mass_matrix = {("x",): self.inv_mass}

        # Provide an initial guess for the parameters
        init_params = {"x": self.model.build_params_array()}

        # Run MCMC with the HMC sampler and the initial guess
        mcmc_kwargs = {
            "num_samples": self.max_iter,
            "warmup_steps": self.warmup,
            "initial_params": init_params,
            "disable_progbar": not self.progress_bar,
        }
        mcmc_kwargs.update(self.mcmc_kwargs)
        mcmc = pyro_MCMC(hmc_kernel, **mcmc_kwargs)

        mcmc.run(self.model, self.prior)
        self.iteration += self.max_iter

        # Extract posterior samples
        chain = mcmc.get_samples()["x"]

        self.chain = chain

        return self



from functools import lru_cache

import torch


def convolve(image, psf):

    image_fft = torch.fft.rfft2(image, s=image.shape)
    psf_fft = torch.fft.rfft2(psf, s=image.shape)

    convolved_fft = image_fft * psf_fft
    convolved = torch.fft.irfft2(convolved_fft, s=image.shape)
    return torch.roll(
        convolved,
        shifts=(-(psf.shape[0] // 2), -(psf.shape[1] // 2)),
        dims=(0, 1),
    )


@lru_cache(maxsize=32)
def curvature_kernel(dtype, device):
    kernel = torch.tensor(
        [
            [0.0, 1.0, 0.0],
            [1.0, -4.0, 1.0],
            [0.0, 1.0, 0.0],
        ],  # [[1., -2.0, 1.], [-2.0, 4, -2.0], [1.0, -2.0, 1.0]],
        device=device,
        dtype=dtype,
    )
    return kernel



import torch


def window_or(other_origin, self_end, other_end):

    new_origin = torch.minimum(-0.5 * torch.ones_like(other_origin), other_origin)
    new_end = torch.maximum(self_end, other_end)

    return new_origin, new_end


def window_and(other_origin, self_end, other_end):
    new_origin = torch.maximum(-0.5 * torch.ones_like(other_origin), other_origin)
    new_end = torch.minimum(self_end, other_end)

    return new_origin, new_end



from .galaxy_model_object import GalaxyModel
from ..utils.decorators import combine_docstrings
from .psf_model_object import PSFModel
from .mixins import (
    ExponentialMixin,
    iExponentialMixin,
    RadialMixin,
    WedgeMixin,
    RayMixin,
    SuperEllipseMixin,
    FourierEllipseMixin,
    WarpMixin,
)

__all__ = [
    "ExponentialGalaxy",
    "ExponentialPSF",
    "ExponentialSuperEllipse",
    "ExponentialFourierEllipse",
    "ExponentialWarp",
    "ExponentialRay",
    "ExponentialWedge",
]


@combine_docstrings
class ExponentialGalaxy(ExponentialMixin, RadialMixin, GalaxyModel):
    usable = True


@combine_docstrings
class ExponentialPSF(ExponentialMixin, RadialMixin, PSFModel):
    _parameter_specs = {"Ie": {"units": "flux/arcsec^2", "value": 1.0}}
    usable = True


@combine_docstrings
class ExponentialSuperEllipse(ExponentialMixin, RadialMixin, SuperEllipseMixin, GalaxyModel):
    usable = True


@combine_docstrings
class ExponentialFourierEllipse(ExponentialMixin, RadialMixin, FourierEllipseMixin, GalaxyModel):
    usable = True


@combine_docstrings
class ExponentialWarp(ExponentialMixin, RadialMixin, WarpMixin, GalaxyModel):
    usable = True


@combine_docstrings
class ExponentialRay(iExponentialMixin, RayMixin, GalaxyModel):
    usable = True


@combine_docstrings
class ExponentialWedge(iExponentialMixin, WedgeMixin, GalaxyModel):
    usable = True



import torch


def ferrer(R, rout, alpha, beta, I0):
    """
    Modified Ferrer profile.

    Parameters
    ----------
    R : array_like
        Radial distance from the center.
    rout : float
        Outer radius of the profile.
    alpha : float
        Power-law index.
    beta : float
        Exponent for the modified Ferrer function.
    I0 : float
        Central intensity.

    Returns
    -------
    array_like
        The modified Ferrer profile evaluated at R.
    """
    return torch.where(
        R < rout,
        I0 * ((1 - (torch.clamp(R, 0, rout) / rout) ** (2 - beta)) ** alpha),
        torch.zeros_like(R),
    )



from .model_object import ComponentModel
from .mixins import InclinedMixin


__all__ = ["GalaxyModel"]


class GalaxyModel(InclinedMixin, ComponentModel):
    """Intended to represent a galaxy or extended component in an image."""

    _model_type = "galaxy"
    usable = False



import numpy as np
import torch

from ..utils.decorators import ignore_numpy_warnings
from .sky_model_object import SkyModel
from ..param import forward

__all__ = ["FlatSky"]


class FlatSky(SkyModel):
    """Model for the sky background in which all values across the image
    are the same.

    Parameters:
        I: brightness for the sky, represented as the log of the brightness over pixel scale squared, this is proportional to a surface brightness

    """

    _model_type = "flat"
    _parameter_specs = {
        "I": {"units": "flux/arcsec^2"},
    }
    usable = True

    @torch.no_grad()
    @ignore_numpy_warnings
    def initialize(self):
        super().initialize()

        if self.I.initialized:
            return

        dat = self.target[self.window].data.detach().cpu().numpy().copy()
        self.I.dynamic_value = np.median(dat) / self.target.pixel_area.item()

    @forward
    def brightness(self, x, y, I):
        return torch.ones_like(x) * I



from scipy.linalg import sqrtm
import numpy as np


def polar_decomposition(A):
    # Step 1: Compute symmetric positive-definite matrix P
    M = A.T @ A
    P = sqrtm(M)  # Principal square root of A^T A

    # Step 2: Compute rotation matrix R
    P_inv = np.linalg.inv(P)
    R = A @ P_inv
    return R, P



from functools import lru_cache

from scipy.special import roots_legendre
import torch


@lru_cache(maxsize=32)
def quad_table(order, dtype, device):
    """
    Generate a meshgrid for quadrature points using Legendre-Gauss quadrature.

    Parameters
    ----------
    n : int
        The number of quadrature points in each dimension.
    dtype : torch.dtype
        The desired data type of the tensor.
    device : torch.device
        The device on which to create the tensor.

    Returns
    -------
    Tuple[torch.Tensor, torch.Tensor, torch.Tensor]
        The generated meshgrid as a tuple of Tensors.
    """
    abscissa, weights = roots_legendre(order)

    w = torch.tensor(weights, dtype=dtype, device=device)
    a = torch.tensor(abscissa, dtype=dtype, device=device) / 2.0
    di, dj = torch.meshgrid(a, a, indexing="ij")

    w = torch.outer(w, w) / 4.0
    return di, dj, w



from .image_object import Image, ImageList
from .target_image import TargetImage, TargetImageList
from .sip_image import SIPTargetImage
from .jacobian_image import JacobianImage, JacobianImageList
from .psf_image import PSFImage
from .model_image import ModelImage, ModelImageList
from .window import Window, WindowList


__all__ = (
    "Image",
    "ImageList",
    "TargetImage",
    "TargetImageList",
    "SIPTargetImage",
    "JacobianImage",
    "JacobianImageList",
    "PSFImage",
    "ModelImage",
    "ModelImageList",
    "Window",
    "WindowList",
)



import torch
import numpy as np

from .psf_model_object import PSFModel
from ..utils.decorators import ignore_numpy_warnings
from ..utils.interpolate import interp2d
from .. import config
from ..errors import SpecificationConflict
from ..param import forward
from . import func
from ..utils.initialize import polar_decomposition

__all__ = ["BasisPSF"]


class PixelBasisPSF(PSFModel):
    """point source model which uses multiple images as a basis for the
    PSF as its representation for point sources. Using bilinear interpolation it
    will shift the PSF within a pixel to accurately represent the center
    location of a point source. There is no functional form for this object type
    as any image can be supplied. Bilinear interpolation is very fast and
    accurate for smooth models, so it is possible to do the expensive
    interpolation before optimization and save time.
    """

    _model_type = "basis"
    _parameter_specs = {
        "weights": {"units": "flux"},
        "PA": {"units": "radians", "shape": ()},
        "scale": {"units": "arcsec/grid-unit", "shape": ()},
    }
    usable = True

    def __init__(self, *args, basis="zernike:3", **kwargs):
        """Initialize the PixelBasisPSF model with a basis set of images."""
        super().__init__(*args, **kwargs)
        self.basis = basis

    @property
    def basis(self):
        """The basis set of images used to form the eigen point source."""
        return self._basis

    @basis.setter
    def basis(self, value):
        """Set the basis set of images. If value is None, the basis is initialized to an empty tensor."""
        if value is None:
            raise SpecificationConflict(
                "PixelBasisPSF requires a basis set of images to be provided."
            )
        elif isinstance(value, str) and value.startswith("zernike:"):
            self._basis = value
        else:
            # Transpose since pytorch uses (j, i) indexing when (i, j) is more natural for coordinates
            self._basis = torch.transpose(
                torch.as_tensor(value, dtype=config.DTYPE, device=config.DEVICE), 1, 2
            )

    @torch.no_grad()
    @ignore_numpy_warnings
    def initialize(self):
        super().initialize()
        target_area = self.target[self.window]
        if not self.PA.initialized:
            R, _ = polar_decomposition(self.target.CD.value.detach().cpu().numpy())
            self.PA.value = np.arccos(np.abs(R[0, 0]))
        if not self.scale.initialized:
            self.scale.value = self.target.pixelscale.item()
        if isinstance(self.basis, str) and self.basis.startswith("zernike:"):
            order = int(self.basis.split(":")[1])
            nm = func.zernike_n_m_list(order)
            N = int(
                target_area.data.shape[0] * self.target.pixelscale.item() / self.scale.value.item()
            )
            X, Y = np.meshgrid(
                np.linspace(-1, 1, N) * (N - 1) / N,
                np.linspace(-1, 1, N) * (N - 1) / N,
                indexing="ij",
            )
            R = np.sqrt(X**2 + Y**2)
            Phi = np.arctan2(Y, X)
            basis = []
            for n, m in nm:
                basis.append(func.zernike_n_m_modes(R, Phi, n, m))
            self.basis = np.stack(basis, axis=0)

        if not self.weights.initialized:
            w = np.zeros(self.basis.shape[0])
            w[0] = 1.0
            self.weights.dynamic_value = w

    @forward
    def transform_coordinates(self, x, y, PA, scale):
        x, y = super().transform_coordinates(x, y)
        i, j = func.rotate(-PA, x, y)
        pixel_center = (self.basis.shape[1] - 1) / 2, (self.basis.shape[2] - 1) / 2
        return i / scale + pixel_center[0], j / scale + pixel_center[1]

    @forward
    def brightness(self, x, y, weights):
        x, y = self.transform_coordinates(x, y)
        return torch.sum(torch.vmap(lambda w, b: w * interp2d(b, y, x))(weights, self.basis), dim=0)



import torch
import numpy as np

from .model_object import ComponentModel
from ..utils.decorators import ignore_numpy_warnings
from . import func
from ..param import forward

__all__ = ["EdgeonModel", "EdgeonSech", "EdgeonIsothermal"]


class EdgeonModel(ComponentModel):
    """General Edge-On galaxy model to be subclassed for any specific
    representation such as radial light profile or the structure of
    the galaxy on the sky. Defines an edgeon galaxy as an object with
    a position angle, no inclination information is included.

    """

    _model_type = "edgeon"
    _parameter_specs = {
        "PA": {"units": "radians", "valid": (0, np.pi), "cyclic": True, "shape": ()},
    }
    usable = False

    @torch.no_grad()
    @ignore_numpy_warnings
    def initialize(self):
        super().initialize()
        if self.PA.initialized:
            return
        target_area = self.target[self.window]
        dat = target_area.data.detach().cpu().numpy().copy()
        edge = np.concatenate((dat[:, 0], dat[:, -1], dat[0, :], dat[-1, :]))
        edge_average = np.median(edge)
        dat = dat - edge_average

        x, y = target_area.coordinate_center_meshgrid()
        x = (x - self.center.value[0]).detach().cpu().numpy()
        y = (y - self.center.value[1]).detach().cpu().numpy()
        mu20 = np.median(dat * np.abs(x))
        mu02 = np.median(dat * np.abs(y))
        mu11 = np.median(dat * x * y / np.sqrt(np.abs(x * y)))
        M = np.array([[mu20, mu11], [mu11, mu02]])
        if np.any(np.iscomplex(M)) or np.any(~np.isfinite(M)):
            self.PA.dynamic_value = np.pi / 2
        else:
            self.PA.dynamic_value = (0.5 * np.arctan2(2 * mu11, mu20 - mu02)) % np.pi

    @forward
    def transform_coordinates(self, x, y, PA):
        x, y = super().transform_coordinates(x, y)
        return func.rotate(-(PA + np.pi / 2), x, y)


class EdgeonSech(EdgeonModel):
    """An edgeon profile where the vertical distribution is a sech^2
    profile, subclasses define the radial profile.

    """

    _model_type = "sech2"
    _parameter_specs = {
        "I0": {"units": "flux/arcsec^2", "shape": ()},
        "hs": {"units": "arcsec", "valid": (0, None), "shape": ()},
    }
    usable = False

    @torch.no_grad()
    @ignore_numpy_warnings
    def initialize(self):
        super().initialize()
        if self.I0.initialized and self.hs.initialized:
            return
        target_area = self.target[self.window]
        icenter = target_area.plane_to_pixel(*self.center.value)

        if not self.I0.initialized:
            chunk = target_area.data[
                int(icenter[0]) - 2 : int(icenter[0]) + 2,
                int(icenter[1]) - 2 : int(icenter[1]) + 2,
            ]
            self.I0.dynamic_value = torch.mean(chunk) / self.target.pixel_area
        if not self.hs.initialized:
            self.hs.value = max(self.window.shape) * target_area.pixelscale * 0.1

    @forward
    def brightness(self, x, y, I0, hs):
        x, y = self.transform_coordinates(x, y)
        return I0 * self.radial_model(x) / (torch.cosh((y + self.softening) / hs) ** 2)


class EdgeonIsothermal(EdgeonSech):
    """A self-gravitating locally-isothermal edgeon disk. This comes from
    van der Kruit & Searle 1981.

    """

    _model_type = "isothermal"
    _parameter_specs = {"rs": {"units": "arcsec", "valid": (0, None), "shape": ()}}
    usable = True

    @torch.no_grad()
    @ignore_numpy_warnings
    def initialize(self):
        super().initialize()
        if self.rs.initialized:
            return
        self.rs.value = max(self.window.shape) * self.target.pixelscale * 0.4

    @forward
    def radial_model(self, R, rs):
        Rscaled = torch.abs(R / rs)
        return (
            Rscaled
            * torch.exp(-Rscaled)
            * torch.special.scaled_modified_bessel_k1(Rscaled + self.softening / rs)
        )



from scipy.stats import binned_statistic, iqr
import numpy as np
import torch
from scipy.optimize import minimize

from ..utils.decorators import ignore_numpy_warnings
from .. import config


def _sample_image(
    image,
    transform,
    radius,
    angle=None,
    rad_bins=None,
    angle_range=None,
    cycle=2 * np.pi,
):
    dat = image.data.detach().cpu().numpy().copy()
    # Fill masked pixels
    if image.has_mask:
        mask = image.mask.detach().cpu().numpy()
        dat[mask] = np.median(dat[~mask])
    # Subtract median of edge pixels to avoid effect of nearby sources
    edge = np.concatenate((dat[:, 0], dat[:, -1], dat[0, :], dat[-1, :]))
    dat -= np.median(edge)
    # Get the radius of each pixel relative to object center
    x, y = transform(*image.coordinate_center_meshgrid(), params=())
    R = radius(x, y, params=()).detach().cpu().numpy().flatten()

    if angle_range is not None:
        T = angle(x, y, params=()).detach().cpu().numpy().flatten()
        T = (T - angle_range[0]) % cycle
        CHOOSE = T < (angle_range[1] - angle_range[0])
        R = R[CHOOSE]
        dat = dat.flatten()[CHOOSE]
    raveldat = dat.ravel()

    # Bin fluxes by radius
    if rad_bins is None:
        rad_bins = np.logspace(
            np.log10(R.min() * 0.9 + image.pixelscale / 2), np.log10(R.max() * 1.1), 11
        )
    else:
        rad_bins = np.array(rad_bins)
    I = (
        binned_statistic(R, raveldat, statistic="median", bins=rad_bins)[0]
    ) / image.pixel_area.item()
    sigma = lambda d: iqr(d, rng=[16, 84]) / 2
    S = binned_statistic(R, raveldat, statistic=sigma, bins=rad_bins)[0] / image.pixel_area.item()
    R = (rad_bins[:-1] + rad_bins[1:]) / 2

    # Ensure enough values are positive
    N = np.isfinite(I)
    I[~N] = np.interp(R[~N], R[N], I[N])
    if np.sum(I > 0) <= 3:
        I = np.abs(I)
    N = I > 0
    if not np.all(N):
        I[~N] = np.interp(R[~N], R[N], I[N])
    # Ensure decreasing brightness with radius in outer regions
    for i in range(5, len(I)):
        if I[i] >= I[i - 1]:
            I[i] = I[i - 1] * 0.9
    # Convert to log scale
    S = S / (I * np.log(10))
    I = np.log10(I)
    # Ensure finite
    N = np.isfinite(I)
    if not np.all(N):
        I[~N] = np.interp(R[~N], R[N], I[N])
    N = np.isfinite(S)
    if not np.all(N):
        S[~N] = np.abs(np.interp(R[~N], R[N], S[N]))
    Sm = np.median(S)
    S[S < Sm] = Sm  # remove very small uncertainties

    return R, I, S


# General parametric
######################################################################
@torch.no_grad()
@ignore_numpy_warnings
def parametric_initialize(model, target, prof_func, params, x0_func):
    if all(list(model[param].initialized for param in params)):
        return
    # Get the sub-image area corresponding to the model image
    R, I, S = _sample_image(target, model.transform_coordinates, model.radius_metric)

    x0 = list(x0_func(model, R, I))
    for i, param in enumerate(params):
        x0[i] = x0[i] if not model[param].initialized else model[param].npvalue

    def optim(x, r, f, u):
        residual = ((f - np.nan_to_num(np.log10(prof_func(r, *x)), nan=np.min(f))) / u) ** 2
        N = np.argsort(residual)
        return np.mean(residual[N][:-2])

    res = minimize(optim, x0=x0, args=(R, I, S), method="Nelder-Mead")

    if res.success:
        x0 = res.x

    for param, x0x in zip(params, x0):
        if not model[param].initialized:
            if not model[param].is_valid(x0x):
                print("soft valid", param, x0x)
                x0x = model[param].soft_valid(
                    torch.tensor(x0x, dtype=config.DTYPE, device=config.DEVICE)
                )
            model[param].dynamic_value = x0x


@torch.no_grad()
@ignore_numpy_warnings
def parametric_segment_initialize(
    model=None,
    target=None,
    prof_func=None,
    params=None,
    x0_func=None,
    segments=None,
):
    if all(list(model[param].initialized for param in params)):
        return

    cycle = np.pi if model.symmetric else 2 * np.pi
    w = cycle / segments
    v = w * np.arange(segments)
    values = []
    for s in range(segments):
        angle_range = (v[s] - w / 2, v[s] + w / 2)
        # Get the sub-image area corresponding to the model image
        R, I, S = _sample_image(
            target,
            model.transform_coordinates,
            model.radius_metric,
            angle=model.angular_metric,
            angle_range=angle_range,
            cycle=cycle,
        )

        x0 = list(x0_func(model, R, I))

        def optim(x, r, f, u):
            residual = ((f - np.log10(prof_func(r, *x))) / u) ** 2
            N = np.argsort(residual)
            return np.mean(residual[N][:-2])

        res = minimize(optim, x0=x0, args=(R, I, S), method="Nelder-Mead")
        if res.success:
            x0 = res.x

        values.append(x0)
    values = np.stack(values).T
    for param, v in zip(params, values):
        if not model[param].initialized:
            model[param].dynamic_value = v



from typing import List, Union

import torch

from .image_object import Image, ImageList
from ..errors import SpecificationConflict, InvalidImage

__all__ = ["JacobianImage", "JacobianImageList"]


######################################################################
class JacobianImage(Image):
    """Jacobian of a model evaluated in an image.

    Image object which represents the evaluation of a jacobian on an
    image. It takes the form of a 3D (Image x Nparameters)
    tensor. This object can be added other other Jacobian images to
    build up a full jacobian for a complex model.

    """

    def __init__(
        self,
        parameters: List[str],
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.parameters = list(parameters)
        if len(self.parameters) != len(set(self.parameters)):
            raise SpecificationConflict("Every parameter should be unique upon jacobian creation")

    def copy(self, **kwargs):
        return super().copy(parameters=self.parameters, **kwargs)

    def match_parameters(self, other: Union["JacobianImage", "JacobianImageList", List]):
        self_i = []
        other_i = []
        other_parameters = other if isinstance(other, list) else other.parameters
        for i, other_param in enumerate(other_parameters):
            if other_param in self.parameters:
                self_i.append(self.parameters.index(other_param))
                other_i.append(i)
        return self_i, other_i

    def __iadd__(self, other: "JacobianImage"):
        if not isinstance(other, JacobianImage):
            raise InvalidImage("Jacobian images can only add with each other, not: type(other)")

        self_indices = self.get_indices(other.window)
        other_indices = other.get_indices(self.window)
        for self_i, other_i in zip(*self.match_parameters(other)):
            self._data[self_indices[0], self_indices[1], self_i] += other.data[
                other_indices[0], other_indices[1], other_i
            ]
        return self


######################################################################
class JacobianImageList(ImageList):
    """For joint modelling, represents Jacobians evaluated on a list of
    images.

    Stores jacobians evaluated on a number of image objects. Since
    jacobian images are aware of the target images they were evaluated
    on, it is possible to combine this object with other
    Jacobian_Image_List objects or even Jacobian_Image objects and
    everything will be sorted into the proper locations of the list,
    and image.

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not all(isinstance(image, (JacobianImage, JacobianImageList)) for image in self.images):
            raise InvalidImage(
                f"JacobianImageList can only hold JacobianImage objects, not {tuple(type(image) for image in self.images)}"
            )

    @property
    def parameters(self) -> List[str]:
        """List of parameters for the jacobian images in this list."""
        if not self.images:
            return []
        return self.images[0].parameters

    def flatten(self, attribute="data"):
        if len(self.images) > 1:
            for image in self.images[1:]:
                if self.images[0].parameters != image.parameters:
                    raise SpecificationConflict(
                        "Jacobian image list sub-images track different parameters. Please initialize with all parameters that will be used."
                    )
        return torch.cat(tuple(image.flatten(attribute) for image in self.images), dim=0)

    def match_parameters(self, other: Union[JacobianImage, "JacobianImageList", List[str]]):
        self_i = []
        other_i = []
        other_parameters = other if isinstance(other, list) else other.parameters
        for i, other_param in enumerate(other_parameters):
            if other_param in self.parameters:
                self_i.append(self.parameters.index(other_param))
                other_i.append(i)
        return self_i, other_i



from .galaxy_model_object import GalaxyModel
from .psf_model_object import PSFModel
from .mixins import (
    FerrerMixin,
    RadialMixin,
    WedgeMixin,
    RayMixin,
    SuperEllipseMixin,
    FourierEllipseMixin,
    WarpMixin,
    iFerrerMixin,
)
from ..utils.decorators import combine_docstrings


__all__ = (
    "FerrerGalaxy",
    "FerrerPSF",
    "FerrerSuperEllipse",
    "FerrerFourierEllipse",
    "FerrerWarp",
    "FerrerRay",
    "FerrerWedge",
)


@combine_docstrings
class FerrerGalaxy(FerrerMixin, RadialMixin, GalaxyModel):
    usable = True


@combine_docstrings
class FerrerPSF(FerrerMixin, RadialMixin, PSFModel):
    _parameter_specs = {"I0": {"units": "flux/arcsec^2", "value": 1.0}}
    usable = True


@combine_docstrings
class FerrerSuperEllipse(FerrerMixin, SuperEllipseMixin, RadialMixin, GalaxyModel):
    usable = True


@combine_docstrings
class FerrerFourierEllipse(FerrerMixin, FourierEllipseMixin, RadialMixin, GalaxyModel):
    usable = True


@combine_docstrings
class FerrerWarp(FerrerMixin, WarpMixin, RadialMixin, GalaxyModel):
    usable = True


@combine_docstrings
class FerrerRay(iFerrerMixin, RayMixin, GalaxyModel):
    usable = True


@combine_docstrings
class FerrerWedge(iFerrerMixin, WedgeMixin, GalaxyModel):
    usable = True



import torch
import numpy as np

from .model_object import ComponentModel
from ..utils.decorators import ignore_numpy_warnings
from . import func
from ..param import forward

__all__ = ["GaussianEllipsoid"]


class GaussianEllipsoid(ComponentModel):
    """Model that represents a galaxy as a 3D Gaussian ellipsoid.

    The model is triaxial, meaning it has three different standard deviations
    along the three axes. The orientation of the ellipsoid is defined by Euler
    angles.

    If all three Euler angles are set to zero, the ellipsoid is aligned with the
    image axes meaning sigma_a gives the std along the x axis of the tangent
    plane, sigma_b gives the std along the y axis of the tangent plane, and
    sigma_z gives the std into the tangent plane. We use the ZXZ convention for
    the Euler angles. This means that for a disk galaxy, one can naturally
    consider sigma_c as the disk thickness and sigma_a=sigma_b as the disk
    radius; setting the Euler angles to zero would leave the disk face-on in the
    x-y tangent plane.

    Note:
        the model is highly degenerate, meaning that it is not possible to
        uniquely determine the parameters from the data. The model is useful if
        one already has a 3D model of the galaxy in mind and wants to produce
        mock data. Alternately, if one applies some constraints on the
        parameters, such as sigma_a = sigma_b and alpha=0, then the model will
        be better determined. In that case, beta is related to the inclination
        of the disk and gamma is related to the position angle of the disk. The
        initialization for this model assumes exactly this interpretation with a
        disk thickness of sigma_c = 0.2 *sigma_a.

    """

    _model_type = "gaussianellipsoid"
    _parameter_specs = {
        "sigma_a": {"units": "arcsec", "valid": (0, None), "shape": ()},
        "sigma_b": {"units": "arcsec", "valid": (0, None), "shape": ()},
        "sigma_c": {"units": "arcsec", "valid": (0, None), "shape": ()},
        "alpha": {"units": "radians", "valid": (0, 2 * np.pi), "cyclic": True, "shape": ()},
        "beta": {"units": "radians", "valid": (0, 2 * np.pi), "cyclic": True, "shape": ()},
        "gamma": {"units": "radians", "valid": (0, 2 * np.pi), "cyclic": True, "shape": ()},
        "flux": {"units": "flux", "shape": ()},
    }
    usable = True

    @torch.no_grad()
    @ignore_numpy_warnings
    def initialize(self):
        super().initialize()

        if any(self[key].initialized for key in GaussianEllipsoid._parameter_specs):
            return

        self.sigma_b = self.sigma_a
        self.sigma_c = lambda p: 0.2 * p.sigma_a.value
        self.sigma_c.link(self.sigma_a)
        self.alpha = 0.0

        target_area = self.target[self.window]
        dat = target_area.data.detach().cpu().numpy().copy()
        if target_area.has_mask:
            mask = target_area.mask.detach().cpu().numpy()
            dat[mask] = np.median(dat[~mask])
        edge = np.concatenate((dat[:, 0], dat[:, -1], dat[0, :], dat[-1, :]))
        edge_average = np.nanmedian(edge)
        dat -= edge_average
        x, y = target_area.coordinate_center_meshgrid()
        center = self.center.value
        x = x - center[0]
        y = y - center[1]
        r = self.radius_metric(x, y, params=()).detach().cpu().numpy()
        self.sigma_a.dynamic_value = np.sqrt(np.sum((r * dat) ** 2) / np.sum(r**2))

        x = x.detach().cpu().numpy()
        y = y.detach().cpu().numpy()

        mu20 = np.median(dat * np.abs(x))
        mu02 = np.median(dat * np.abs(y))
        mu11 = np.median(dat * x * y / np.sqrt(np.abs(x * y) + self.softening**2))
        M = np.array([[mu20, mu11], [mu11, mu02]])
        if np.any(np.iscomplex(M)) or np.any(~np.isfinite(M)):
            PA = np.pi / 2
            l = (0.7, 1.0)
        else:
            PA = (0.5 * np.arctan2(2 * mu11, mu20 - mu02) - np.pi / 2) % np.pi
            l = np.sort(np.linalg.eigvals(M))
        q = np.clip(np.sqrt(l[0] / l[1]), 0.1, 0.9)
        self.beta.dynamic_value = np.arccos(q)
        self.gamma.dynamic_value = PA
        self.flux.dynamic_value = np.sum(dat)

    @forward
    def brightness(self, x, y, sigma_a, sigma_b, sigma_c, alpha, beta, gamma, flux):
        """Brightness of the Gaussian ellipsoid."""
        D = torch.diag(torch.stack((sigma_a, sigma_b, sigma_c)) ** 2)
        R = func.euler_rotation_matrix(alpha, beta, gamma)
        Sigma = R @ D @ R.T
        Sigma2D = Sigma[:2, :2]
        inv_Sigma = torch.linalg.inv(Sigma2D)
        v = torch.stack(self.transform_coordinates(x, y), dim=0).reshape(2, -1)
        return (
            flux
            * torch.exp(-0.5 * (v * (inv_Sigma @ v)).sum(dim=0))
            / (2 * np.pi * torch.linalg.det(Sigma2D).sqrt())
        ).reshape(x.shape)



from typing import Optional, Union

import torch
import numpy as np
from astropy.wcs import WCS as AstropyWCS
from astropy.io import fits

from ..param import Module, Param, forward
from .. import config
from ..utils.conversions.units import deg_to_arcsec, arcsec_to_deg
from .window import Window, WindowList
from ..errors import InvalidImage, SpecificationConflict

# from .base import BaseImage
from . import func

__all__ = ["Image", "ImageList"]


class Image(Module):
    """Core class to represent images with pixel values, pixel scale,
    and a window defining the spatial coordinates on the sky.
    It supports arithmetic operations with other image objects while preserving logical image boundaries.
    It also provides methods for determining the coordinate locations of pixels
    """

    default_CD = ((1.0, 0.0), (0.0, 1.0))
    expect_ctype = (("RA---TAN",), ("DEC--TAN",))

    def __init__(
        self,
        *,
        data: Optional[torch.Tensor] = None,
        CD: Optional[Union[float, torch.Tensor]] = None,
        zeropoint: Optional[Union[float, torch.Tensor]] = None,
        crpix: Union[torch.Tensor, tuple] = (0.0, 0.0),
        crtan: Union[torch.Tensor, tuple] = (0.0, 0.0),
        crval: Union[torch.Tensor, tuple] = (0.0, 0.0),
        pixelscale: Optional[Union[torch.Tensor, float]] = None,
        wcs: Optional[AstropyWCS] = None,
        filename: Optional[str] = None,
        hduext=0,
        identity: str = None,
        name: Optional[str] = None,
        _data: Optional[torch.Tensor] = None,
    ) -> None:
        """Initialize an instance of the APImage class.

        Parameters:
        -----------
        data : numpy.ndarray or None, optional
            The image data. Default is None.
        wcs : astropy.wcs.wcs.WCS or None, optional
            A WCS object which defines a coordinate system for the image. Note that AstroPhot only handles basic WCS conventions. It will use the WCS object to get `wcs.pixel_to_world(-0.5, -0.5)` to determine the position of the origin in world coordinates. It will also extract the `pixel_scale_matrix` to index pixels going forward.
        pixelscale : float or None, optional
            The physical scale of the pixels in the image, in units of arcseconds. Default is None.
        filename : str or None, optional
            The name of a file containing the image data. Default is None.
        zeropoint : float or None, optional
            The image's zeropoint, used for flux calibration. Default is None.

        """
        super().__init__(name=name)
        if _data is None:
            self.data = data  # units: flux
        else:
            self._data = _data
        self.crval = Param(
            "crval", shape=(2,), units="deg", dtype=config.DTYPE, device=config.DEVICE
        )
        self.crtan = Param(
            "crtan",
            crtan,
            shape=(2,),
            units="arcsec",
            dtype=config.DTYPE,
            device=config.DEVICE,
        )
        self.CD = Param(
            "CD",
            shape=(2, 2),
            units="arcsec/pixel",
            dtype=config.DTYPE,
            device=config.DEVICE,
        )
        self.zeropoint = zeropoint

        if filename is not None:
            self.load(filename, hduext=hduext)
            return

        if identity is None:
            self.identity = id(self)
        else:
            self.identity = identity

        if wcs is not None:
            if wcs.wcs.ctype[0] not in self.expect_ctype[0]:
                config.logger.warning(
                    "Astropy WCS not tangent plane coordinate system! May not be compatible with AstroPhot."
                )
            if wcs.wcs.ctype[1] not in self.expect_ctype[1]:
                config.logger.warning(
                    "Astropy WCS not tangent plane coordinate system! May not be compatible with AstroPhot."
                )

            crval = wcs.wcs.crval
            crpix = np.array(wcs.wcs.crpix)[::-1] - 1  # handle FITS 1-indexing

            if CD is not None:
                config.logger.warning("WCS CD set with supplied WCS, ignoring user supplied CD!")
            CD = deg_to_arcsec * wcs.pixel_scale_matrix

        # set the data
        self.crval = crval
        self.crpix = crpix

        if isinstance(CD, (float, int)):
            CD = np.array([[CD, 0.0], [0.0, CD]], dtype=np.float64)
        elif CD is None and pixelscale is not None:
            CD = np.array([[pixelscale, 0.0], [0.0, pixelscale]], dtype=np.float64)
        elif CD is None:
            CD = self.default_CD
        self.CD = CD

    @property
    def data(self):
        """The image data, which is a tensor of pixel values."""
        return self._data

    @data.setter
    def data(self, value: Optional[torch.Tensor]):
        """Set the image data. If value is None, the data is initialized to an empty tensor."""
        if value is None:
            self._data = torch.empty((0, 0), dtype=config.DTYPE, device=config.DEVICE)
        else:
            # Transpose since pytorch uses (j, i) indexing when (i, j) is more natural for coordinates
            self._data = torch.transpose(
                torch.as_tensor(value, dtype=config.DTYPE, device=config.DEVICE), 0, 1
            )

    @property
    def crpix(self):
        """The reference pixel coordinates in the image, which is used to convert from pixel coordinates to tangent plane coordinates."""
        return self._crpix

    @crpix.setter
    def crpix(self, value: Union[torch.Tensor, tuple]):
        self._crpix = np.asarray(value, dtype=np.float64)

    @property
    def zeropoint(self):
        """The zeropoint of the image, which is used to convert from pixel flux to magnitude."""
        return self._zeropoint

    @zeropoint.setter
    def zeropoint(self, value):
        """Set the zeropoint of the image."""
        if value is None:
            self._zeropoint = None
        else:
            self._zeropoint = torch.as_tensor(value, dtype=config.DTYPE, device=config.DEVICE)

    @property
    def window(self):
        return Window(window=((0, 0), self.data.shape[:2]), image=self)

    @property
    def center(self):
        shape = torch.as_tensor(self.data.shape[:2], dtype=config.DTYPE, device=config.DEVICE)
        return torch.stack(self.pixel_to_plane(*((shape - 1) / 2)))

    @property
    def shape(self):
        """The shape of the image data."""
        return self.data.shape

    @property
    @forward
    def pixel_area(self, CD):
        """The area inside a pixel in arcsec^2"""
        return torch.linalg.det(CD).abs()

    @property
    @forward
    def pixelscale(self):
        """The approximate side length of a pixel, which is just
        sqrt(pixel_area). For square pixels this is the actual pixel
        length, for rectangular pixels it is a kind of average.

        The pixelscale is not used for exact calculations
        and instead sets a size scale within an image.

        """
        return self.pixel_area.sqrt()

    @forward
    def pixel_to_plane(self, i, j, crtan, CD):
        return func.pixel_to_plane_linear(i, j, *self.crpix, CD, *crtan)

    @forward
    def plane_to_pixel(self, x, y, crtan, CD):
        return func.plane_to_pixel_linear(x, y, *self.crpix, CD, *crtan)

    @forward
    def plane_to_world(self, x, y, crval):
        return func.plane_to_world_gnomonic(x, y, *crval)

    @forward
    def world_to_plane(self, ra, dec, crval):
        return func.world_to_plane_gnomonic(ra, dec, *crval)

    @forward
    def world_to_pixel(self, ra, dec):
        """A wrapper which applies :meth:`world_to_plane` then
        :meth:`plane_to_pixel`, see those methods for further
        information.

        """
        return self.plane_to_pixel(*self.world_to_plane(ra, dec))

    @forward
    def pixel_to_world(self, i, j):
        """A wrapper which applies :meth:`pixel_to_plane` then
        :meth:`plane_to_world`, see those methods for further
        information.

        """
        return self.plane_to_world(*self.pixel_to_plane(i, j))

    def pixel_center_meshgrid(self):
        """Get a meshgrid of pixel coordinates in the image, centered on the pixel grid."""
        return func.pixel_center_meshgrid(self.shape, config.DTYPE, config.DEVICE)

    def pixel_corner_meshgrid(self):
        """Get a meshgrid of pixel coordinates in the image, with corners at the pixel grid."""
        return func.pixel_corner_meshgrid(self.shape, config.DTYPE, config.DEVICE)

    def pixel_simpsons_meshgrid(self):
        """Get a meshgrid of pixel coordinates in the image, with Simpson's rule sampling."""
        return func.pixel_simpsons_meshgrid(self.shape, config.DTYPE, config.DEVICE)

    def pixel_quad_meshgrid(self, order=3):
        """Get a meshgrid of pixel coordinates in the image, with quadrature sampling."""
        return func.pixel_quad_meshgrid(self.shape, config.DTYPE, config.DEVICE, order=order)

    @forward
    def coordinate_center_meshgrid(self):
        """Get a meshgrid of coordinate locations in the image, centered on the pixel grid."""
        i, j = self.pixel_center_meshgrid()
        return self.pixel_to_plane(i, j)

    @forward
    def coordinate_corner_meshgrid(self):
        """Get a meshgrid of coordinate locations in the image, with corners at the pixel grid."""
        i, j = self.pixel_corner_meshgrid()
        return self.pixel_to_plane(i, j)

    @forward
    def coordinate_simpsons_meshgrid(self):
        """Get a meshgrid of coordinate locations in the image, with Simpson's rule sampling."""
        i, j = self.pixel_simpsons_meshgrid()
        return self.pixel_to_plane(i, j)

    @forward
    def coordinate_quad_meshgrid(self, order=3):
        """Get a meshgrid of coordinate locations in the image, with quadrature sampling."""
        i, j, _ = self.pixel_quad_meshgrid(order=order)
        return self.pixel_to_plane(i, j)

    def copy(self, **kwargs):
        """Produce a copy of this image with all of the same properties. This
        can be used when one wishes to make temporary modifications to
        an image and then will want the original again.

        """
        kwargs = {
            "_data": torch.clone(self.data.detach()),
            "CD": self.CD.value,
            "crpix": self.crpix,
            "crval": self.crval.value,
            "crtan": self.crtan.value,
            "zeropoint": self.zeropoint,
            "identity": self.identity,
            "name": self.name,
            **kwargs,
        }
        return self.__class__(**kwargs)

    def blank_copy(self, **kwargs):
        """Produces a blank copy of the image which has the same properties
        except that its data is now filled with zeros.

        """
        kwargs = {
            "_data": torch.zeros_like(self.data),
            **kwargs,
        }
        return self.copy(**kwargs)

    def crop(self, pixels, **kwargs):
        """Crop the image by the number of pixels given. This will crop
        the image in all four directions by the number of pixels given.

        given data shape (N, M) the new shape will be:

        crop - int: crop the same number of pixels on all sides. new shape (N - 2*crop, M - 2*crop)
        crop - (int, int): crop each dimension by the number of pixels given. new shape (N - 2*crop[1], M - 2*crop[0])
        crop - (int, int, int, int): crop each side by the number of pixels given assuming (x low, x high, y low, y high). new shape (N - crop[2] - crop[3], M - crop[0] - crop[1])
        """
        if isinstance(pixels, int):
            data = self.data[
                pixels : self.data.shape[0] - pixels,
                pixels : self.data.shape[1] - pixels,
            ]
            crpix = self.crpix - pixels
        elif len(pixels) == 1:  # same crop in all dimension
            crop = pixels if isinstance(pixels, int) else pixels[0]
            data = self.data[
                crop : self.data.shape[0] - crop,
                crop : self.data.shape[1] - crop,
            ]
            crpix = self.crpix - crop
        elif len(pixels) == 2:  # different crop in each dimension
            data = self.data[
                pixels[0] : self.data.shape[0] - pixels[0],
                pixels[1] : self.data.shape[1] - pixels[1],
            ]
            crpix = self.crpix - pixels
        elif len(pixels) == 4:  # different crop on all sides
            data = self.data[
                pixels[0] : self.data.shape[0] - pixels[1],
                pixels[2] : self.data.shape[1] - pixels[3],
            ]
            crpix = self.crpix - pixels[0::2]
        else:
            raise ValueError(
                f"Invalid crop shape {pixels}, must be (int,), (int, int), or (int, int, int, int)!"
            )
        return self.copy(_data=data, crpix=crpix, **kwargs)

    def reduce(self, scale: int, **kwargs):
        """This operation will downsample an image by the factor given. If
        scale = 2 then 2x2 blocks of pixels will be summed together to
        form individual larger pixels. A new image object will be
        returned with the appropriate pixelscale and data tensor. Note
        that the window does not change in this operation since the
        pixels are condensed, but the pixel size is increased
        correspondingly.

        Parameters:
            scale: factor by which to condense the image pixels. Each scale X scale region will be summed [int]

        """
        if not isinstance(scale, int) and not (
            isinstance(scale, torch.Tensor) and scale.dtype is torch.int32
        ):
            raise SpecificationConflict(f"Reduce scale must be an integer! not {type(scale)}")
        if scale == 1:
            return self

        MS = self.data.shape[0] // scale
        NS = self.data.shape[1] // scale

        data = self.data[: MS * scale, : NS * scale].reshape(MS, scale, NS, scale).sum(axis=(1, 3))
        CD = self.CD.value * scale
        crpix = (self.crpix + 0.5) / scale - 0.5
        return self.copy(
            _data=data,
            CD=CD,
            crpix=crpix,
            **kwargs,
        )

    def to(self, dtype=None, device=None):
        if dtype is None:
            dtype = config.DTYPE
        if device is None:
            device = config.DEVICE
        super().to(dtype=dtype, device=device)
        self._data = self._data.to(dtype=dtype, device=device)
        if self.zeropoint is not None:
            self.zeropoint = self.zeropoint.to(dtype=dtype, device=device)
        return self

    def flatten(self, attribute: str = "data") -> torch.Tensor:
        return getattr(self, attribute).flatten(end_dim=1)

    def fits_info(self):
        return {
            "CTYPE1": "RA---TAN",
            "CTYPE2": "DEC--TAN",
            "CRVAL1": self.crval.value[0].item(),
            "CRVAL2": self.crval.value[1].item(),
            "CRPIX1": self.crpix[0] + 1,
            "CRPIX2": self.crpix[1] + 1,
            "CRTAN1": self.crtan.value[0].item(),
            "CRTAN2": self.crtan.value[1].item(),
            "CD1_1": self.CD.value[0][0].item() * arcsec_to_deg,
            "CD1_2": self.CD.value[0][1].item() * arcsec_to_deg,
            "CD2_1": self.CD.value[1][0].item() * arcsec_to_deg,
            "CD2_2": self.CD.value[1][1].item() * arcsec_to_deg,
            "MAGZP": self.zeropoint.item() if self.zeropoint is not None else -999,
            "IDNTY": self.identity,
        }

    def fits_images(self):
        return [
            fits.PrimaryHDU(
                torch.transpose(self.data, 0, 1).detach().cpu().numpy(),
                header=fits.Header(self.fits_info()),
            )
        ]

    def get_astropywcs(self, **kwargs):
        kwargs = {
            "NAXIS": 2,
            "NAXIS1": self.shape[0].item(),
            "NAXIS2": self.shape[1].item(),
            **self.fits_info(),
            **kwargs,
        }
        return AstropyWCS(kwargs)

    def save(self, filename: str):
        hdulist = fits.HDUList(self.fits_images())
        hdulist.writeto(filename, overwrite=True)

    def load(self, filename: str, hduext=0):
        """Load an image from a FITS file. This will load the primary HDU
        and set the data, CD, crpix, crval, and crtan attributes
        accordingly. If the WCS is not tangent plane, it will warn the user.

        """
        hdulist = fits.open(filename)
        self.data = np.array(hdulist[hduext].data, dtype=np.float64)

        self.CD = (
            np.array(
                (
                    (hdulist[hduext].header["CD1_1"], hdulist[hduext].header["CD1_2"]),
                    (hdulist[hduext].header["CD2_1"], hdulist[hduext].header["CD2_2"]),
                ),
                dtype=np.float64,
            )
            * deg_to_arcsec
        )
        self.crpix = (hdulist[hduext].header["CRPIX1"] - 1, hdulist[hduext].header["CRPIX2"] - 1)
        self.crval = (hdulist[hduext].header["CRVAL1"], hdulist[hduext].header["CRVAL2"])
        if "CRTAN1" in hdulist[hduext].header and "CRTAN2" in hdulist[hduext].header:
            self.crtan = (hdulist[hduext].header["CRTAN1"], hdulist[hduext].header["CRTAN2"])
        if "MAGZP" in hdulist[hduext].header and hdulist[hduext].header["MAGZP"] > -998:
            self.zeropoint = hdulist[hduext].header["MAGZP"]
        self.identity = hdulist[hduext].header.get("IDNTY", str(id(self)))
        return hdulist

    def corners(self):
        pixel_lowleft = torch.tensor((-0.5, -0.5), dtype=config.DTYPE, device=config.DEVICE)
        pixel_lowright = torch.tensor(
            (self.data.shape[0] - 0.5, -0.5), dtype=config.DTYPE, device=config.DEVICE
        )
        pixel_upleft = torch.tensor(
            (-0.5, self.data.shape[1] - 0.5), dtype=config.DTYPE, device=config.DEVICE
        )
        pixel_upright = torch.tensor(
            (self.data.shape[0] - 0.5, self.data.shape[1] - 0.5),
            dtype=config.DTYPE,
            device=config.DEVICE,
        )
        lowleft = self.pixel_to_plane(*pixel_lowleft)
        lowright = self.pixel_to_plane(*pixel_lowright)
        upleft = self.pixel_to_plane(*pixel_upleft)
        upright = self.pixel_to_plane(*pixel_upright)
        return (lowleft, lowright, upright, upleft)

    @torch.no_grad()
    def get_indices(self, other: Window):
        if other.image is self:
            return slice(max(0, other.i_low), min(self.shape[0], other.i_high)), slice(
                max(0, other.j_low), min(self.shape[1], other.j_high)
            )
        shift = np.round(self.crpix - other.crpix).astype(int)
        return slice(
            min(max(0, other.i_low + shift[0]), self.shape[0]),
            max(0, min(other.i_high + shift[0], self.shape[0])),
        ), slice(
            min(max(0, other.j_low + shift[1]), self.shape[1]),
            max(0, min(other.j_high + shift[1], self.shape[1])),
        )

    @torch.no_grad()
    def get_other_indices(self, other: Window):
        if other.image == self:
            shape = other.shape
            return slice(max(0, -other.i_low), min(self.shape[0] - other.i_low, shape[0])), slice(
                max(0, -other.j_low), min(self.shape[1] - other.j_low, shape[1])
            )
        raise ValueError()

    def get_window(self, other: Union[Window, "Image"], indices=None, **kwargs):
        """Get a new image object which is a window of this image
        corresponding to the other image's window. This will return a
        new image object with the same properties as this one, but with
        the data cropped to the other image's window.

        """
        if indices is None:
            indices = self.get_indices(other if isinstance(other, Window) else other.window)
        new_img = self.copy(
            _data=self.data[indices],
            crpix=self.crpix - np.array((indices[0].start, indices[1].start)),
            **kwargs,
        )
        return new_img

    def __sub__(self, other):
        if isinstance(other, Image):
            new_img = self[other]
            new_img._data = new_img.data - other[self].data
            return new_img
        else:
            new_img = self.copy()
            new_img._data = new_img.data - other
            return new_img

    def __add__(self, other):
        if isinstance(other, Image):
            new_img = self[other]
            new_img._data = new_img.data + other[self].data
            return new_img
        else:
            new_img = self.copy()
            new_img._data = new_img.data + other
            return new_img

    def __iadd__(self, other):
        if isinstance(other, Image):
            self._data[self.get_indices(other.window)] += other.data[other.get_indices(self.window)]
        else:
            self._data = self.data + other
        return self

    def __isub__(self, other):
        if isinstance(other, Image):
            self._data[self.get_indices(other.window)] -= other.data[other.get_indices(self.window)]
        else:
            self._data = self.data - other
        return self

    def __getitem__(self, *args):
        if len(args) == 1 and isinstance(args[0], (Image, Window)):
            return self.get_window(args[0])
        return super().__getitem__(*args)


class ImageList(Module):
    def __init__(self, images, name=None):
        super().__init__(name=name)
        self.images = list(images)
        if not all(isinstance(image, Image) for image in self.images):
            raise InvalidImage(
                f"Image_List can only hold Image objects, not {tuple(type(image) for image in self.images)}"
            )

    @property
    def data(self):
        return tuple(image.data for image in self.images)

    def copy(self):
        return self.__class__(
            tuple(image.copy() for image in self.images),
        )

    def blank_copy(self):
        return self.__class__(
            tuple(image.blank_copy() for image in self.images),
        )

    def get_window(self, other: "ImageList"):
        return self.__class__(
            tuple(image[win] for image, win in zip(self.images, other.images)),
        )

    def index(self, other: Image):
        for i, image in enumerate(self.images):
            if other.identity == image.identity:
                return i
        else:
            raise IndexError(
                f"Could not find identity match between image list {self.name} and input image {other.name}"
            )

    def match_indices(self, other: "ImageList"):
        """Match the indices of the images in this list with those in another Image_List."""
        indices = []
        for other_image in other.images:
            try:
                i = self.index(other_image)
            except IndexError:
                continue
            indices.append(i)
        return indices

    def to(self, dtype=None, device=None):
        if dtype is not None:
            dtype = config.DTYPE
        if device is not None:
            device = config.DEVICE
        super().to(dtype=dtype, device=device)
        return self

    def flatten(self, attribute="data"):
        return torch.cat(tuple(image.flatten(attribute) for image in self.images))

    def __sub__(self, other):
        if isinstance(other, ImageList):
            new_list = []
            for other_image in other.images:
                i = self.index(other_image)
                self_image = self.images[i]
                new_list.append(self_image - other_image)
            return self.__class__(new_list)
        else:
            raise ValueError("Subtraction of Image_List only works with another Image_List object!")

    def __add__(self, other):
        if isinstance(other, ImageList):
            new_list = []
            for other_image in other.images:
                try:
                    i = self.index(other_image)
                except IndexError:
                    continue
                self_image = self.images[i]
                new_list.append(self_image + other_image)
            return self.__class__(new_list)
        else:
            raise ValueError("Addition of Image_List only works with another Image_List object!")

    def __isub__(self, other):
        if isinstance(other, ImageList):
            for other_image in other.images:
                try:
                    i = self.index(other_image)
                except IndexError:
                    continue
                self.images[i] -= other_image
        elif isinstance(other, Image):
            i = self.index(other)
            self.images[i] -= other
        else:
            raise ValueError("Subtraction of Image_List only works with another Image_List object!")
        return self

    def __iadd__(self, other):
        if isinstance(other, ImageList):
            for other_image in other.images:
                try:
                    i = self.index(other_image)
                except IndexError:
                    continue
                self.images[i] += other_image
        elif isinstance(other, Image):
            i = self.index(other)
            self.images[i] += other
        else:
            raise ValueError("Addition of Image_List only works with another Image_List object!")
        return self

    def __getitem__(self, *args):
        if len(args) == 1:
            if isinstance(args[0], ImageList):
                new_list = []
                for other_image in args[0].images:
                    i = self.index(other_image)
                    new_list.append(self.images[i].get_window(other_image))
                return self.__class__(new_list)
            elif isinstance(args[0], WindowList):
                new_list = []
                for other_window in args[0].windows:
                    i = self.index(other_window.image)
                    new_list.append(self.images[i].get_window(other_window))
                return self.__class__(new_list)
            elif isinstance(args[0], Image):
                i = self.index(args[0])
                return self.images[i].get_window(args[0])
            elif isinstance(args[0], Window):
                i = self.index(args[0].image)
                return self.images[i].get_window(args[0])
            elif isinstance(args[0], int):
                return self.images[args[0]]
        super().__getitem__(*args)

    def __iter__(self):
        return (img for img in self.images)



from .galaxy_model_object import GalaxyModel
from .psf_model_object import PSFModel
from .mixins import (
    KingMixin,
    RadialMixin,
    WedgeMixin,
    RayMixin,
    SuperEllipseMixin,
    FourierEllipseMixin,
    WarpMixin,
    iKingMixin,
)
from ..utils.decorators import combine_docstrings


__all__ = (
    "KingGalaxy",
    "KingPSF",
    "KingSuperEllipse",
    "KingFourierEllipse",
    "KingWarp",
    "KingRay",
    "KingWedge",
)


@combine_docstrings
class KingGalaxy(KingMixin, RadialMixin, GalaxyModel):
    usable = True


@combine_docstrings
class KingPSF(KingMixin, RadialMixin, PSFModel):
    _parameter_specs = {"I0": {"units": "flux/arcsec^2", "value": 1.0}}
    usable = True


@combine_docstrings
class KingSuperEllipse(KingMixin, SuperEllipseMixin, RadialMixin, GalaxyModel):
    usable = True


@combine_docstrings
class KingFourierEllipse(KingMixin, FourierEllipseMixin, RadialMixin, GalaxyModel):
    usable = True


@combine_docstrings
class KingWarp(KingMixin, WarpMixin, RadialMixin, GalaxyModel):
    usable = True


@combine_docstrings
class KingRay(iKingMixin, RayMixin, GalaxyModel):
    usable = True


@combine_docstrings
class KingWedge(iKingMixin, WedgeMixin, GalaxyModel):
    usable = True



from .galaxy_model_object import GalaxyModel

from .psf_model_object import PSFModel
from .mixins import (
    GaussianMixin,
    RadialMixin,
    WedgeMixin,
    RayMixin,
    SuperEllipseMixin,
    FourierEllipseMixin,
    WarpMixin,
    iGaussianMixin,
)
from ..utils.decorators import combine_docstrings


__all__ = [
    "GaussianGalaxy",
    "GaussianPSF",
    "GaussianSuperEllipse",
    "GaussianFourierEllipse",
    "GaussianWarp",
    "GaussianRay",
    "GaussianWedge",
]


@combine_docstrings
class GaussianGalaxy(GaussianMixin, RadialMixin, GalaxyModel):
    usable = True


@combine_docstrings
class GaussianPSF(GaussianMixin, RadialMixin, PSFModel):
    _parameter_specs = {"flux": {"units": "flux", "value": 1.0}}
    usable = True


@combine_docstrings
class GaussianSuperEllipse(GaussianMixin, SuperEllipseMixin, RadialMixin, GalaxyModel):
    usable = True


@combine_docstrings
class GaussianFourierEllipse(GaussianMixin, FourierEllipseMixin, RadialMixin, GalaxyModel):
    usable = True


@combine_docstrings
class GaussianWarp(GaussianMixin, WarpMixin, RadialMixin, GalaxyModel):
    usable = True


@combine_docstrings
class GaussianRay(iGaussianMixin, RayMixin, GalaxyModel):
    usable = True


@combine_docstrings
class GaussianWedge(iGaussianMixin, WedgeMixin, GalaxyModel):
    usable = True



import torch
import numpy as np

from ...utils.integration import quad_table


def pixel_center_integrator(Z: torch.Tensor):
    return Z


def pixel_corner_integrator(Z: torch.Tensor):
    kernel = torch.ones((1, 1, 2, 2), dtype=Z.dtype, device=Z.device) / 4.0
    Z = torch.nn.functional.conv2d(Z.view(1, 1, *Z.shape), kernel, padding="valid")
    return Z.squeeze(0).squeeze(0)


def pixel_simpsons_integrator(Z: torch.Tensor):
    kernel = (
        torch.tensor([[[[1, 4, 1], [4, 16, 4], [1, 4, 1]]]], dtype=Z.dtype, device=Z.device) / 36.0
    )
    Z = torch.nn.functional.conv2d(Z.view(1, 1, *Z.shape), kernel, padding="valid", stride=2)
    return Z.squeeze(0).squeeze(0)


def pixel_quad_integrator(Z: torch.Tensor, w: torch.Tensor = None, order=3):
    """
    Integrate the pixel values using quadrature weights.

    Parameters
    ----------
    Z : torch.Tensor
        The tensor containing pixel values.
    w : torch.Tensor
        The quadrature weights.

    Returns
    -------
    torch.Tensor
        The integrated value.
    """
    if w is None:
        _, _, w = quad_table(order, Z.dtype, Z.device)
    Z = Z * w
    return Z.sum(dim=(-1))


def upsample(i, j, order, scale):
    dp = torch.linspace(-1, 1, order, dtype=i.dtype, device=i.device) * (order - 1) / (2.0 * order)
    di, dj = torch.meshgrid(dp, dp, indexing="xy")

    si = torch.repeat_interleave(i.unsqueeze(-1), order**2, -1) + scale * di.flatten()
    sj = torch.repeat_interleave(j.unsqueeze(-1), order**2, -1) + scale * dj.flatten()
    return si, sj


def single_quad_integrate(i, j, brightness_ij, scale, quad_order=3):
    di, dj, w = quad_table(quad_order, i.dtype, i.device)
    qi = torch.repeat_interleave(i.unsqueeze(-1), quad_order**2, -1) + scale * di.flatten()
    qj = torch.repeat_interleave(j.unsqueeze(-1), quad_order**2, -1) + scale * dj.flatten()
    z = brightness_ij(qi, qj)
    z0 = torch.mean(z, dim=-1)
    z = torch.sum(z * w.flatten(), dim=-1)
    return z, z0


def recursive_quad_integrate(
    i,
    j,
    brightness_ij,
    threshold,
    scale=1.0,
    quad_order=3,
    gridding=5,
    _current_depth=0,
    max_depth=1,
):
    scale = 1 / (gridding**_current_depth)
    z, z0 = single_quad_integrate(i, j, brightness_ij, scale, quad_order)

    if _current_depth >= max_depth:
        return z

    select = torch.abs(z - z0) > threshold / scale**2

    integral = torch.zeros_like(z)
    integral[~select] = z[~select]

    si, sj = upsample(i[select], j[select], quad_order, scale)

    integral[select] = recursive_quad_integrate(
        si,
        sj,
        brightness_ij,
        threshold,
        scale=scale,
        quad_order=quad_order,
        gridding=gridding,
        _current_depth=_current_depth + 1,
        max_depth=max_depth,
    ).mean(dim=-1)

    return integral


def recursive_bright_integrate(
    i,
    j,
    brightness_ij,
    bright_frac,
    scale=1.0,
    quad_order=3,
    gridding=5,
    _current_depth=0,
    max_depth=1,
):
    scale = 1 / (gridding**_current_depth)
    z, _ = single_quad_integrate(i, j, brightness_ij, scale, quad_order)

    if _current_depth >= max_depth:
        return z

    N = max(1, int(np.prod(z.shape) * bright_frac))
    z_flat = z.flatten()

    select = torch.topk(z_flat, N, dim=-1).indices

    si, sj = upsample(i.flatten()[select], j.flatten()[select], quad_order, scale)

    z_flat[select] = recursive_bright_integrate(
        si,
        sj,
        brightness_ij,
        bright_frac,
        scale=scale,
        quad_order=quad_order,
        gridding=gridding,
        _current_depth=_current_depth + 1,
        max_depth=max_depth,
    ).mean(dim=-1)

    return z_flat.reshape(z.shape)



from typing import Literal
import numpy as np
import torch

from astropy.visualization import HistEqStretch, ImageNormalize
from matplotlib.patches import Polygon
import matplotlib
from scipy.stats import iqr

from ..models import GroupModel, PSFModel, PSFGroupModel
from ..image import ImageList, WindowList
from .. import config
from ..utils.conversions.units import flux_to_sb
from ..utils.decorators import ignore_numpy_warnings
from .visuals import *


__all__ = ["target_image", "psf_image", "model_image", "residual_image", "model_window"]


@ignore_numpy_warnings
def target_image(fig, ax, target, window=None, **kwargs):
    """
    This function is used to display a target image using the provided figure and axes.

    Args:
        fig (matplotlib.figure.Figure): The figure object in which the target image will be displayed.
        ax (matplotlib.axes.Axes): The axes object on which the target image will be plotted.
        target (Image or Image_List): The image or list of images to be displayed.
        window (Window, optional): The window through which the image is viewed. If `None`, the window of the
            provided `target` is used. Defaults to `None`.
        **kwargs: Arbitrary keyword arguments.

    Returns:
        fig (matplotlib.figure.Figure): The figure object containing the displayed target image.
        ax (matplotlib.axes.Axes): The axes object containing the displayed target image.

    Note:
        If the `target` is an `Image_List`, this function will recursively call itself for each image in the list.
        The `window` parameter and `kwargs` are passed unchanged to each recursive call.
    """

    # recursive call for target image list
    if isinstance(target, ImageList):
        for i in range(len(target.images)):
            target_image(fig, ax[i], target.images[i], window=window, **kwargs)
        return fig, ax
    if window is None:
        window = target.window
    target_area = target[window]
    dat = np.copy(target_area.data.detach().cpu().numpy())
    if target_area.has_mask:
        dat[target_area.mask.detach().cpu().numpy()] = np.nan
    X, Y = target_area.coordinate_corner_meshgrid()
    X = X.detach().cpu().numpy()
    Y = Y.detach().cpu().numpy()
    sky = np.nanmedian(dat)
    noise = iqr(dat[np.isfinite(dat)], rng=(16, 84)) / 2
    if noise == 0:
        noise = np.nanstd(dat)
    vmin = sky - 5 * noise
    vmax = sky + 5 * noise

    if kwargs.get("linear", False):
        im = ax.pcolormesh(
            X,
            Y,
            dat,
            cmap=cmap_grad,
        )
    else:
        im = ax.pcolormesh(
            X,
            Y,
            dat,
            cmap="gray_r",
            norm=ImageNormalize(
                stretch=HistEqStretch(
                    dat[np.logical_and(dat <= (sky + 3 * noise), np.isfinite(dat))]
                ),
                clip=False,
                vmax=sky + 3 * noise,
                vmin=np.nanmin(dat),
            ),
        )
        pickhist = dat < (sky + 3 * noise)
        if np.sum(~pickhist) > 5:  # only draw log if multiple pixels above noise
            im = ax.pcolormesh(
                X,
                Y,
                np.ma.masked_where(pickhist, dat),
                cmap=cmap_grad,
                norm=matplotlib.colors.LogNorm(),
                clim=[sky + 3 * noise, None],
            )

    if torch.linalg.det(target.CD.value) < 0:
        ax.invert_xaxis()
    ax.axis("equal")
    ax.set_xlabel("Tangent Plane X [arcsec]")
    ax.set_ylabel("Tangent Plane Y [arcsec]")

    return fig, ax


@torch.no_grad()
@ignore_numpy_warnings
def psf_image(
    fig,
    ax,
    psf,
    cmap_levels=None,
    vmin=None,
    vmax=None,
    **kwargs,
):
    if isinstance(psf, (PSFModel, PSFGroupModel)):
        psf = psf()
    # recursive call for target image list
    if isinstance(psf, ImageList):
        for i in range(len(psf.images)):
            psf_image(fig, ax[i], psf.images[i], **kwargs)
        return fig, ax

    # Evaluate the model image
    x, y = psf.coordinate_corner_meshgrid()
    x = x.detach().cpu().numpy()
    y = y.detach().cpu().numpy()
    psf = psf.data.detach().cpu().numpy()

    # Default kwargs for image
    kwargs = {
        "cmap": cmap_grad,
        "norm": matplotlib.colors.LogNorm(
            vmin=vmin, vmax=vmax
        ),  # "norm": ImageNormalize(stretch=LogStretch(), clip=False),
        **kwargs,
    }

    # if requested, convert the continuous colourmap into discrete levels
    if cmap_levels is not None:
        kwargs["cmap"] = matplotlib.colors.ListedColormap(
            list(kwargs["cmap"](c) for c in np.linspace(0.0, 1.0, cmap_levels))
        )

    # Plot the image
    ax.pcolormesh(x, y, psf, **kwargs)

    # Enforce equal spacing on x y
    ax.axis("equal")
    ax.set_xlabel("PSF X [arcsec]")
    ax.set_ylabel("PSF Y [arcsec]")

    return fig, ax


@torch.no_grad()
@ignore_numpy_warnings
def model_image(
    fig,
    ax,
    model,
    sample_image=None,
    window=None,
    target=None,
    showcbar=True,
    target_mask=False,
    cmap_levels=None,
    magunits=True,
    **kwargs,
):
    """
    This function is used to generate a model image and display it using the provided figure and axes.

    Args:
        fig (matplotlib.figure.Figure): The figure object in which the image will be displayed.
        ax (matplotlib.axes.Axes): The axes object on which the image will be plotted.
        model (Model): The model object used to generate a model image if `sample_image` is not provided.
        sample_image (Image or Image_List, optional): The image or list of images to be displayed.
            If `None`, a model image is generated using the provided `model`. Defaults to `None`.
        window (Window, optional): The window through which the image is viewed. If `None`, the window of the
            provided `model` is used. Defaults to `None`.
        target (Target, optional): The target or list of targets for the image or image list.
            If `None`, the target of the `model` is used. Defaults to `None`.
        showcbar (bool, optional): Whether to show the color bar. Defaults to `True`.
        target_mask (bool, optional): Whether to apply the mask of the target. If `True` and if the target has a mask,
            the mask is applied to the image. Defaults to `False`.
        cmap_levels (int, optional): The number of discrete levels to convert the continuous color map to.
            If not `None`, the color map is converted to a ListedColormap with the specified number of levels.
            Defaults to `None`.
        **kwargs: Arbitrary keyword arguments. These are used to override the default imshow_kwargs.

    Returns:
        fig (matplotlib.figure.Figure): The figure object containing the displayed image.
        ax (matplotlib.axes.Axes): The axes object containing the displayed image.

    Note:
        If the `sample_image` is an `Image_List`, this function will recursively call itself for each image in the list,
        with the corresponding target and window. The `showcbar` parameter and `kwargs` are passed unchanged to each recursive call.
    """

    if sample_image is None:
        sample_image = model()

    # Use model target if not given
    if target is None:
        target = model.target

    # Use model window if not given
    if window is None:
        window = model.window

    # Handle image lists
    if isinstance(sample_image, ImageList):
        for i, (images, targets, windows) in enumerate(zip(sample_image, target, window)):
            model_image(
                fig,
                ax[i],
                model,
                sample_image=images,
                window=windows,
                target=targets,
                showcbar=showcbar,
                target_mask=target_mask,
                cmap_levels=cmap_levels,
                magunits=magunits,
                **kwargs,
            )
        return fig, ax

    # cut out the requested window
    sample_image = sample_image[window]

    # Evaluate the model image
    X, Y = sample_image.coordinate_corner_meshgrid()
    X = X.detach().cpu().numpy()
    Y = Y.detach().cpu().numpy()
    sample_image = sample_image.data.detach().cpu().numpy()

    # Default kwargs for image
    kwargs = {
        "cmap": cmap_grad,
        **kwargs,
    }

    # if requested, convert the continuous colourmap into discrete levels
    if cmap_levels is not None:
        kwargs["cmap"] = matplotlib.colors.ListedColormap(
            list(kwargs["cmap"](c) for c in np.linspace(0.0, 1.0, cmap_levels))
        )

    # If zeropoint is available, convert to surface brightness units
    if target.zeropoint is not None and magunits:
        sample_image = flux_to_sb(sample_image, target.pixel_area.item(), target.zeropoint.item())
        kwargs["cmap"] = kwargs["cmap"].reversed()
    else:
        vmin = kwargs.pop("vmin", None)
        vmax = kwargs.pop("vmax", None)
        kwargs = {
            "norm": matplotlib.colors.LogNorm(
                vmin=vmin, vmax=vmax
            ),  # "norm": ImageNormalize(stretch=LogStretch(), clip=False),
            **kwargs,
        }

    # Apply the mask if available
    if target_mask and target.has_mask:
        sample_image[target.mask.detach().cpu().numpy()] = np.nan

    # Plot the image
    im = ax.pcolormesh(X, Y, sample_image, **kwargs)

    if torch.linalg.det(target.CD.value) < 0:
        ax.invert_xaxis()

    # Enforce equal spacing on x y
    ax.axis("equal")
    ax.set_xlabel("Tangent Plane X [arcsec]")
    ax.set_ylabel("Tangent Plane Y [arcsec]")

    # Add a colourbar
    if showcbar:
        if target.zeropoint is not None and magunits:
            clb = fig.colorbar(im, ax=ax, label="Surface Brightness [mag/arcsec$^2$]")
            clb.ax.invert_yaxis()
        else:
            clb = fig.colorbar(im, ax=ax, label="log$_{10}$(flux)")

    return fig, ax


@torch.no_grad()
@ignore_numpy_warnings
def residual_image(
    fig,
    ax,
    model,
    target=None,
    sample_image=None,
    showcbar=True,
    window=None,
    clb_label=None,
    normalize_residuals=False,
    scaling: Literal["arctan", "clip", "none"] = "arctan",
    **kwargs,
):
    """
    This function is used to calculate and display the residuals of a model image with respect to a target image.
    The residuals are calculated as the difference between the target image and the sample image.

    Args:
        fig (matplotlib.figure.Figure): The figure object in which the residuals will be displayed.
        ax (matplotlib.axes.Axes): The axes object on which the residuals will be plotted.
        model (Model): The model object used to generate a model image if `sample_image` is not provided.
        target (Target or Image_List, optional): The target or list of targets for the image or image list.
            If `None`, the target of the `model` is used. Defaults to `None`.
        sample_image (Image or Image_List, optional): The image or list of images from which residuals will be calculated.
            If `None`, a model image is generated using the provided `model`. Defaults to `None`.
        showcbar (bool, optional): Whether to show the color bar. Defaults to `True`.
        window (Window or Window_List, optional): The window through which the image is viewed. If `None`, the window of the
            provided `model` is used. Defaults to `None`.
        center_residuals (bool, optional): Whether to subtract the median of the residuals. If `True`, the median is subtracted
            from the residuals. Defaults to `False`.
        clb_label (str, optional): The label for the colorbar. If `None`, a default label is used based on the normalization of the
            residuals. Defaults to `None`.
        normalize_residuals (bool, optional): Whether to normalize the residuals. If `True`, residuals are divided by the square root
            of the variance of the target. Defaults to `False`.
        sample_full_image: If True, every model will be sampled on the full image window. If False (default) each model will only be sampled in its fitting window.
        **kwargs: Arbitrary keyword arguments. These are used to override the default imshow_kwargs.

    Returns:
        fig (matplotlib.figure.Figure): The figure object containing the displayed residuals.
        ax (matplotlib.axes.Axes): The axes object containing the displayed residuals.

    Note:
        If the `window`, `target`, or `sample_image` are lists, this function will recursively call itself for each element in the list,
        with the corresponding window, target, and sample image. The `showcbar`, `center_residuals`, and `kwargs` are passed unchanged to
        each recursive call.
    """

    if window is None:
        window = model.window
    if target is None:
        target = model.target
    if sample_image is None:
        sample_image = model()
    if isinstance(window, WindowList) or isinstance(target, ImageList):
        for i_ax, win, tar, sam in zip(ax, window, target, sample_image):
            residual_image(
                fig,
                i_ax,
                model,
                target=tar,
                sample_image=sam,
                window=win,
                showcbar=showcbar,
                clb_label=clb_label,
                normalize_residuals=normalize_residuals,
                **kwargs,
            )
        return fig, ax

    sample_image = sample_image[window]
    target = target[window]
    X, Y = sample_image.coordinate_corner_meshgrid()
    X = X.detach().cpu().numpy()
    Y = Y.detach().cpu().numpy()
    residuals = (target - sample_image).data

    if normalize_residuals is True:
        residuals = residuals / torch.sqrt(target.variance)
    elif isinstance(normalize_residuals, torch.Tensor):
        residuals = residuals / torch.sqrt(normalize_residuals)
        normalize_residuals = True
    if target.has_mask:
        residuals[target.mask] = np.nan
    residuals = residuals.detach().cpu().numpy()

    if scaling == "clip":
        if normalize_residuals is not True:
            config.logger.warning(
                "Using clipping scaling without normalizing residuals. This may lead to confusing results."
            )
        residuals = np.clip(residuals, -5, 5)
        vmax = 5
        default_label = (
            f"(Target - {model.name}) / $\\sigma$"
            if normalize_residuals
            else f"(Target - {model.name})"
        )
    elif scaling == "arctan":
        residuals = np.arctan(
            residuals / (iqr(residuals[np.isfinite(residuals)], rng=[10, 90]) * 2)
        )
        vmax = np.pi / 2
        if normalize_residuals:
            default_label = f"tan$^{{-1}}$((Target - {model.name}) / $\\sigma$)"
        else:
            default_label = f"tan$^{{-1}}$(Target - {model.name})"
    elif scaling == "none":
        vmax = np.max(np.abs(residuals[np.isfinite(residuals)]))
        default_label = (
            f"(Target - {model.name}) / $\\sigma$"
            if normalize_residuals
            else f"(Target - {model.name})"
        )
    else:
        raise ValueError(f"Unknown scaling type {scaling}. Use 'clip', 'arctan', or 'none'.")
    imshow_kwargs = {
        "cmap": cmap_div,
        "vmin": -vmax,
        "vmax": vmax,
    }
    imshow_kwargs.update(kwargs)
    im = ax.pcolormesh(X, Y, residuals, **imshow_kwargs)
    if torch.linalg.det(target.CD.value) < 0:
        ax.invert_xaxis()
    ax.axis("equal")
    ax.set_xlabel("Tangent Plane X [arcsec]")
    ax.set_ylabel("Tangent Plane Y [arcsec]")

    if showcbar:
        clb = fig.colorbar(im, ax=ax, label=default_label if clb_label is None else clb_label)
        clb.ax.set_yticks([])
        clb.ax.set_yticklabels([])
    return fig, ax


@ignore_numpy_warnings
def model_window(fig, ax, model, target=None, rectangle_linewidth=2, **kwargs):
    if target is None:
        target = model.target
    if isinstance(ax, np.ndarray):
        for i, axitem in enumerate(ax):
            model_window(fig, axitem, model, target=target.images[i], **kwargs)
        return fig, ax

    if isinstance(model, GroupModel):
        for m in model.models:
            if isinstance(m.window, WindowList):
                use_window = m.window.windows[m.target.index(target)]
            else:
                use_window = m.window

            corners = target[use_window].corners()
            x = [
                corners[0][0].item(),
                corners[1][0].item(),
                corners[2][0].item(),
                corners[3][0].item(),
            ]
            y = [
                corners[0][1].item(),
                corners[1][1].item(),
                corners[2][1].item(),
                corners[3][1].item(),
            ]
            ax.add_patch(
                Polygon(
                    xy=list(zip(x, y)),
                    fill=False,
                    linewidth=rectangle_linewidth,
                    edgecolor=main_pallet["secondary1"],
                )
            )
    else:
        use_window = model.window
        corners = target[use_window].corners()
        x = [
            corners[0][0].item(),
            corners[1][0].item(),
            corners[2][0].item(),
            corners[3][0].item(),
        ]
        y = [
            corners[0][1].item(),
            corners[1][1].item(),
            corners[2][1].item(),
            corners[3][1].item(),
        ]
        ax.add_patch(
            Polygon(
                xy=list(zip(x, y)),
                fill=False,
                linewidth=rectangle_linewidth,
                edgecolor=main_pallet["secondary1"],
            )
        )

    return fig, ax



from ..param import forward
from .galaxy_model_object import GalaxyModel
from .psf_model_object import PSFModel
from ..utils.conversions.functions import sersic_Ie_to_flux_torch
from ..utils.decorators import combine_docstrings
from .mixins import (
    SersicMixin,
    RadialMixin,
    WedgeMixin,
    iSersicMixin,
    RayMixin,
    SuperEllipseMixin,
    FourierEllipseMixin,
    WarpMixin,
    TruncationMixin,
)

__all__ = [
    "SersicGalaxy",
    "TSersicGalaxy",
    "SersicPSF",
    "Sersic_Warp",
    "Sersic_SuperEllipse",
    "Sersic_FourierEllipse",
    "Sersic_Ray",
    "Sersic_Wedge",
]


@combine_docstrings
class SersicGalaxy(SersicMixin, RadialMixin, GalaxyModel):
    usable = True

    @forward
    def total_flux(self, Ie, n, Re, q, window=None):
        return sersic_Ie_to_flux_torch(Ie, n, Re, q)


@combine_docstrings
class TSersicGalaxy(TruncationMixin, SersicMixin, RadialMixin, GalaxyModel):
    usable = True


@combine_docstrings
class SersicPSF(SersicMixin, RadialMixin, PSFModel):
    _parameter_specs = {"Ie": {"units": "flux/arcsec^2", "value": 1.0}}
    usable = True

    @forward
    def total_flux(self, Ie, n, Re):
        return sersic_Ie_to_flux_torch(Ie, n, Re, 1.0)


@combine_docstrings
class SersicSuperEllipse(SersicMixin, RadialMixin, SuperEllipseMixin, GalaxyModel):
    usable = True


@combine_docstrings
class SersicFourierEllipse(SersicMixin, RadialMixin, FourierEllipseMixin, GalaxyModel):
    usable = True


@combine_docstrings
class SersicWarp(SersicMixin, RadialMixin, WarpMixin, GalaxyModel):
    usable = True


@combine_docstrings
class SersicRay(iSersicMixin, RayMixin, GalaxyModel):
    usable = True


@combine_docstrings
class SersicWedge(iSersicMixin, WedgeMixin, GalaxyModel):
    usable = True



import numpy as np
import torch
from scipy.special import gamma
from torch.special import gammaln


def sersic_n_to_b(n):
    """Compute the `b(n)` for a sersic model. This factor ensures that
    the :math:`R_e` and :math:`I_e` parameters do in fact correspond
    to the half light values and not some other scale
    radius/intensity.

    """

    return (
        2 * n
        - 1 / 3
        + 4 / (405 * n)
        + 46 / (25515 * n**2)
        + 131 / (1148175 * n**3)
        - 2194697 / (30690717750 * n**4)
    )


def sersic_I0_to_flux_np(I0, n, R, q):
    """Compute the total flux integrated to infinity for a 2D elliptical
    sersic given the :math:`I_0,n,R_s,q` parameters which uniquely
    define the profile (:math:`I_0` is the central intensity in
    flux/arcsec^2). Note that :math:`R_s` is not the effective radius,
    but in fact the scale radius in the more straightforward sersic
    representation:

    .. math::

      I(R) = I_0e^{-(R/R_s)^{1/n}}

    Args:
      I0: central intensity (flux/arcsec^2)
      n: sersic index
      R: Scale radius
      q: axis ratio (b/a)

    """
    return 2 * np.pi * I0 * q * n * R**2 * gamma(2 * n)


def sersic_flux_to_I0_np(flux, n, R, q):
    """Compute the central intensity (flux/arcsec^2) for a 2D elliptical
    sersic given the :math:`F,n,R_s,q` parameters which uniquely
    define the profile (:math:`F` is the total flux integrated to
    infinity). Note that :math:`R_s` is not the effective radius, but
    in fact the scale radius in the more straightforward sersic
    representation:

    .. math::

      I(R) = I_0e^{-(R/R_s)^{1/n}}

    Args:
      flux: total flux integrated to infinity (flux)
      n: sersic index
      R: Scale radius
      q: axis ratio (b/a)

    """
    return flux / (2 * np.pi * q * n * R**2 * gamma(2 * n))


def sersic_Ie_to_flux_np(Ie, n, R, q):
    """Compute the total flux integrated to infinity for a 2D elliptical
    sersic given the :math:`I_e,n,R_e,q` parameters which uniquely
    define the profile (:math:`I_e` is the intensity at :math:`R_e` in
    flux/arcsec^2). Note that :math:`R_e` is the effective radius in
    the sersic representation:

    .. math::

      I(R) = I_ee^{-b_n[(R/R_e)^{1/n}-1]}

    Args:
      Ie: intensity at the effective radius (flux/arcsec^2)
      n: sersic index
      R: Scale radius
      q: axis ratio (b/a)

    """
    bn = sersic_n_to_b(n)
    return 2 * np.pi * Ie * R**2 * q * n * (np.exp(bn) * bn ** (-2 * n)) * gamma(2 * n)


def sersic_flux_to_Ie_np(flux, n, R, q):
    """Compute the intensity at :math:`R_e` (flux/arcsec^2) for a 2D
    elliptical sersic given the :math:`F,n,R_e,q` parameters which
    uniquely define the profile (:math:`F` is the total flux
    integrated to infinity). Note that :math:`R_e` is the effective
    radius in the sersic representation:

    .. math::

      I(R) = I_ee^{-b_n[(R/R_e)^{1/n}-1]}

    Args:
      flux: flux integrated to infinity (flux)
      n: sersic index
      R: Scale radius
      q: axis ratio (b/a)

    """
    bn = sersic_n_to_b(n)
    return flux / (2 * np.pi * R**2 * q * n * (np.exp(bn) * bn ** (-2 * n)) * gamma(2 * n))


def sersic_inv_np(I, n, Re, Ie):
    """Invert the sersic profile. Compute the radius corresponding to a
    given intensity for a pure sersic profile.

    """
    bn = sersic_n_to_b(n)
    return Re * ((1 - (1 / bn) * np.log(I / Ie)) ** (n))


def sersic_I0_to_flux_torch(I0, n, R, q):
    """Compute the total flux integrated to infinity for a 2D elliptical
    sersic given the :math:`I_0,n,R_s,q` parameters which uniquely
    define the profile (:math:`I_0` is the central intensity in
    flux/arcsec^2). Note that :math:`R_s` is not the effective radius,
    but in fact the scale radius in the more straightforward sersic
    representation:

    .. math::

      I(R) = I_0e^{-(R/R_s)^{1/n}}

    Args:
      I0: central intensity (flux/arcsec^2)
      n: sersic index
      R: Scale radius
      q: axis ratio (b/a)


    """
    return 2 * np.pi * I0 * q * n * R**2 * torch.exp(gammaln(2 * n))


def sersic_flux_to_I0_torch(flux, n, R, q):
    """Compute the central intensity (flux/arcsec^2) for a 2D elliptical
    sersic given the :math:`F,n,R_s,q` parameters which uniquely
    define the profile (:math:`F` is the total flux integrated to
    infinity). Note that :math:`R_s` is not the effective radius, but
    in fact the scale radius in the more straightforward sersic
    representation:

    .. math::

      I(R) = I_0e^{-(R/R_s)^{1/n}}

    Args:
      flux: total flux integrated to infinity (flux)
      n: sersic index
      R: Scale radius
      q: axis ratio (b/a)


    """
    return flux / (2 * np.pi * q * n * R**2 * torch.exp(gammaln(2 * n)))


def sersic_Ie_to_flux_torch(Ie, n, R, q):
    """Compute the total flux integrated to infinity for a 2D elliptical
    sersic given the :math:`I_e,n,R_e,q` parameters which uniquely
    define the profile (:math:`I_e` is the intensity at :math:`R_e` in
    flux/arcsec^2). Note that :math:`R_e` is the effective radius in
    the sersic representation:

    .. math::

      I(R) = I_ee^{-b_n[(R/R_e)^{1/n}-1]}

    Args:
      Ie: intensity at the effective radius (flux/arcsec^2)
      n: sersic index
      R: Scale radius
      q: axis ratio (b/a)


    """
    bn = sersic_n_to_b(n)
    return (
        2 * np.pi * Ie * R**2 * q * n * (torch.exp(bn) * bn ** (-2 * n)) * torch.exp(gammaln(2 * n))
    )


def sersic_flux_to_Ie_torch(flux, n, R, q):
    """Compute the intensity at :math:`R_e` (flux/arcsec^2) for a 2D
    elliptical sersic given the :math:`F,n,R_e,q` parameters which
    uniquely define the profile (:math:`F` is the total flux
    integrated to infinity). Note that :math:`R_e` is the effective
    radius in the sersic representation:

    .. math::

      I(R) = I_ee^{-b_n[(R/R_e)^{1/n}-1]}

    Args:
      flux: flux integrated to infinity (flux)
      n: sersic index
      R: Scale radius
      q: axis ratio (b/a)


    """
    bn = sersic_n_to_b(n)
    return flux / (
        2 * np.pi * R**2 * q * n * (torch.exp(bn) * bn ** (-2 * n)) * torch.exp(gammaln(2 * n))
    )


def sersic_inv_torch(I, n, Re, Ie):
    """Invert the sersic profile. Compute the radius corresponding to a
    given intensity for a pure sersic profile.

    """
    bn = sersic_n_to_b(n)
    return Re * ((1 - (1 / bn) * torch.log(I / Ie)) ** (n))


def moffat_I0_to_flux(I0, n, rd, q):
    """
    Compute the total flux integrated to infinity for a moffat profile.

    Args:
      I0: central intensity (flux/arcsec^2)
      n: moffat curvature parameter (unitless)
      rd: scale radius
      q: axis ratio
    """
    return I0 * np.pi * rd**2 * q / (n - 1)
